# -*- coding: utf-8 -*-

# stdlib
import unittest

# pypi
from pyramid import testing

# local
from pyramid_session_redis.connection import get_default_connection
from . import DummyRedis


# ==============================================================================


class TestConnection(unittest.TestCase):
    def setUp(self):
        testing.setUp(self)
        self.request = testing.DummyRequest()

    def tearDown(self):
        testing.tearDown(self)

    def test_get_default_connection(self):
        options = dict(host="localhost", port=999)
        inst = get_default_connection(
            self.request, url=None, client_class=DummyRedis, **options
        )  # noqa: E501
        self.assertEqual(inst.host, "localhost")
        self.assertEqual(inst.port, 999)

    def test_get_default_connection_with_url(self):
        url = "redis://username:password@localhost:6379/0"
        inst = get_default_connection(
            self.request, url=url, client_class=DummyRedis
        )  # noqa: E501
        self.assertEqual(inst.url, url)

    def test_get_default_connection_url_removes_duplicates(self):
        options = dict(host="localhost", port=999, password="password", db=5)
        url = "redis://username:password@localhost:6379/0"
        inst = get_default_connection(
            self.request, url=url, client_class=DummyRedis, **options
        )
        self.assertEqual(inst.url, url)
        self.assertNotIn("password", inst.opts)
        self.assertNotIn("host", inst.opts)
        self.assertNotIn("port", inst.opts)
        self.assertNotIn("db", inst.opts)



# -*- coding: utf-8 -*-

# stdlib
import datetime
from enum import Enum
from functools import partial
from math import ceil
from secrets import token_urlsafe
from time import time as time_time
import typing
from typing import Any
from typing import Callable
from typing import Dict
from typing import Optional
from typing import Type
from typing import TYPE_CHECKING
from typing import Union
import warnings

# pypi
from pyramid.exceptions import ConfigurationError
from pyramid.settings import asbool
from redis.exceptions import WatchError
from typing_extensions import Protocol

# local
from .exceptions import InvalidSessionId_Deserialization
from .exceptions import InvalidSessionId_Serialization

if TYPE_CHECKING:
    from .session import RedisSession
    from redis.client import Redis as RedisClient

    # from webob.cookies import _Serializer

# ==============================================================================


# we use an Enum for typing support


# create a custom class+object instance for handling lazycreated ids
# (this is what dogpile cache's NO_VALUE does)
class LazyCreateSession(Enum):
    pass


# used to differentiate from `None`
class NotSpecified(Enum):
    pass


# this stored in the sessions. it is used to detect api version mismatches
SESSION_API_VERSION: int = 1


TYPING_COOKIE_EXPIRES = Union[datetime.timedelta, datetime.datetime, None, NotSpecified]
TYPING_KEY = Union[str, int]
TYPING_SESSION_ID = Union[str, Type[LazyCreateSession]]

# for ASSIGNMENT we need a `Type[NotSpecified]`
TYPING_COOKIE_EXPIRES__A = Union[
    datetime.timedelta, datetime.datetime, None, Type[NotSpecified]
]
TYPING_COOKIE_MAX_AGE__A = Union[int, None, Type[NotSpecified]]


# ------------------------------------------------------------------------------


def warn_future(message: str) -> None:
    warnings.warn(message, FutureWarning, stacklevel=2)


def is_integer(n: typing.Any) -> bool:
    try:
        r = float(n)
        return r.is_integer()
    except ValueError:
        return False


def to_binary(value: typing.AnyStr, enc: str = "UTF-8") -> bytes:
    if isinstance(value, str):
        return value.encode(enc)
    return value


def _generate_session_id() -> str:
    """
    Produces a base64 encoded, urlsafe random string with 48-byte
    cryptographically strong randomness as the session id. See

        http://security.stackexchange.com/questions/24850/
        choosing-a-session-id-algorithm-for-a-client-server-relationship

    for the algorithm of choosing a session id.

    The implementation of `os.urandom` varies by system, but you can always
    supply your own function in your ini file with:

        redis.sessions.id_generator = my_random_id_generator

    This uses 48 bytes instead of 32 to maintain backwards
    compatibility to pyramid_redis_sessions.  The earlier packaged used
    a 64character digest; however 48bits using the new method will
    encode to a 64 character url-safe string, while 32 bits will only be
    encoded to a 40 character string.
    """
    return token_urlsafe(48)


# ---------------------

# `_parse_settings` and `includeme` may need to coerce strings into other types
# these lists are maintained here as a public API, so implementers who need
# to customize their integration can do so without fear of breaking changes

# these we just ignore in code, but track here:
configs_str = (
    "secret",
    "cookie_name",
    "cookie_path",
    "cookie_domain",
    "cookie_comment",
    "cookie_samesite",
    "host",  # DEPRECATED in 1.7; REMOVING in 1.8
    "password",  # DEPRECATED in 1.7; REMOVING in 1.8
    "redis_host",
    "redis_password",
    "redis_unix_socket_path",
    "redis_encoding_errors",
    "redis_encoding",
    "redis_url",
    "url",  # DEPRECATED in 1.7; REMOVING in 1.8
)

# treat as strings here
configs_dotable = (
    "client_callable",  # DEPRECATED in 1.7; REMOVING in 1.8
    "cookie_signer",
    "deserialize",
    "func_check_response_allow_cookies",
    "func_invalid_logger",
    "id_generator",
    "redis_client_callable",
    "serialize",
)

configs_bool = (
    "cookie_httponly",
    "cookie_on_exception",
    "cookie_secure",
    "deserialized_fails_new",
    "detect_changes",
    "invalidate_empty_session",
    "python_expires",
    "set_redis_ttl_readheavy",
    "set_redis_ttl",
)

configs_int = (
    "db",  # DEPRECATED in 1.7; REMOVING in 1.8
    "port",  # DEPRECATED in 1.7; REMOVING in 1.8
    "redis_db",
    "redis_port",
)

configs_int_none = (
    "cookie_max_age",
    "redis_socket_timeout",
    "timeout",
    "timeout_trigger",
)

configs_unsupported = (
    "cookie_expires",  # datetime objects
    "redis_connection_pool",  # ConnectionPool
)


# ---------------------


def prefixed_id(prefix: str = "session:") -> str:
    """
    :param prefix: string. the prefix
    Adds a prefix to the unique session id, for cases where you want to
    visually distinguish keys in redis.
    """
    session_id: str = _generate_session_id()
    prefixed_id: str = prefix + session_id
    return prefixed_id


def int_time() -> int:
    return int(ceil(time_time()))


def empty_session_payload(
    timeout: int = 0,
    python_expires: Optional[bool] = None,
) -> Dict:
    """
    creates an empty session payload

    :param timeout: int. default 0.
    :param python_expires: bool. default None.
    """
    _created = int_time()
    data = {
        "m": {},  # managed_dict
        "c": _created,  # created
        "v": SESSION_API_VERSION,  # session
    }
    if timeout:
        data["t"] = timeout  # timeout
        if python_expires:
            data["x"] = _created + timeout
    return data


def encode_session_payload(
    managed_dict: Dict,
    created: int,
    timeout: int,  # = 0?
    expires: int,  # = 0?
    timeout_trigger: Optional[int] = None,
    python_expires: Optional[bool] = None,
) -> Dict:
    """
    called by a session to recode for storage;
    inverse of ``decode_session_payload``

    :param managed_dict: internal dict for encoding.
    :param created: int. time created.
    :param timeout: int. seconds.
    :param expires: int. time expiry.
    :param timeout_trigger: int. default None.
    :param python_expires: bool. default None.
    """
    data = {
        "m": managed_dict,  # managed_dict
        "c": created,  # created
        "v": SESSION_API_VERSION,  # session_api version
    }
    if expires and python_expires:
        data["x"] = expires
    if timeout:
        data["t"] = timeout  # timeout
        if python_expires:
            time_now = int_time()
            if not timeout_trigger or (time_now >= (expires - timeout_trigger)):
                data["x"] = time_now + timeout  # expires
    return data


def decode_session_payload(payload: Dict) -> Dict:
    """
    decode a serialized session payload to kwargs
    inverse of ``encode_session_payload``

    :param payload: dict with encoding compatible with `encode_session_payload`
    :returns payload: dict with legacy/readble format.
    """
    return {
        "managed_dict": payload["m"],
        "created": payload["c"],
        "version": payload["v"],
        "timeout": payload.get("t"),
        "expires": payload.get("x"),
    }


def _insert_session_id_if_unique(
    redis: "RedisClient",
    timeout: int,
    session_id: str,
    serialize: Callable[[Any], bytes],
    set_redis_ttl: bool,
    data_payload: Optional[dict] = None,
    new_payload_func: Optional[Callable[..., Dict]] = None,
    python_expires: Optional[bool] = None,
) -> Optional[str]:
    """
    Attempt to insert a given ``session_id`` AND empty payload; returning the
    successful id or ``None``.

    ``timeout`` could be 0/None, in that case do-not track
    the timeout data.

    This will create an empty/null session and redis entry for the id.

    The return value will be the input `session_id` upon success, or `None` upon
    a failure.

    ``data_payload`` = payload to use
    ``new_payload_func`` = specify a fallback function to generate a payload
    if both are ``None``, then `empty_session_payload`

    :param redis: redis connection
    :param timeout: int seconds. used to initialize empty session.
    :param session_id: string.
    :param serialize: callable. used to serialize an empty session.
    :param set_redis_ttl:  bool
    :param data_payload: dict. default None. initialize session with this payload.
    :param new_payload_func: callable. default None. create a new payload with this.
    :param python_expires: bool. default None.
    :returns session_id: string.
    """
    if data_payload is None:
        if new_payload_func is not None:
            data_payload = new_payload_func()
        else:
            data_payload = empty_session_payload(
                timeout,
                python_expires=python_expires,
            )
    _payload = serialize(data_payload)
    with redis.pipeline() as pipe:
        try:
            # start pipeline with a watch
            pipe.watch(session_id)
            # after `watch` the pipline is in immediate execution mode
            value = pipe.get(session_id)
            if value is not None:
                return None
            # enter buffered mode
            pipe.multi()
            if timeout and set_redis_ttl:
                pipe.setex(session_id, timeout, _payload)
            else:
                pipe.set(session_id, _payload)
            pipe.execute()
            # if a WatchError wasn't raised during execution, everything
            # we just did happened atomically
            return session_id
        except WatchError:
            return None


def create_unique_session_id(
    redis: "RedisClient",
    timeout: int,
    serialize: Callable[[Any], bytes],
    generator: Callable[[], str] = _generate_session_id,
    set_redis_ttl: bool = True,
    data_payload: Optional[Dict] = None,
    new_payload_func: Optional[Callable[..., Dict]] = None,
    python_expires: Optional[bool] = None,
) -> str:
    """
    Returns a unique session id after inserting it successfully in Redis.

    :param redis: redis connection
    :param timeout: int seconds. used to initialize empty session.
    :param serialize: callable. used to serialize an empty session.
    :param generator: callable. used to generate an id.
    :param set_redis_ttl:  bool
    :param data_payload: dict. default None. initialize session with this payload.
    :param new_payload_func: callable. default None. create a new payload with this.
    :param python_expires: bool. default None.
    :returns:
    """
    while 1:
        session_id: str = generator()
        # attempt will be the session_id
        attempt = _insert_session_id_if_unique(
            redis,
            timeout,
            session_id,
            serialize,
            set_redis_ttl,
            data_payload=data_payload,
            new_payload_func=new_payload_func,
            python_expires=python_expires,
        )
        if attempt is not None:
            return attempt


def _parse_settings(settings: Dict) -> Dict:
    """
    Convenience function to collect settings prefixed by 'redis.sessions' and
    coerce settings to ``int``, ``float``, and ``bool`` as needed.

    :param settings: dict
    """
    keys = [s for s in settings if s.startswith("redis.sessions.")]

    options = {}

    for k in keys:
        param = k.split(".")[-1]
        value = settings[k]
        options[param] = value

    # check unsupported
    _unsupported = []
    for u in configs_unsupported:
        if u in options:
            _unsupported.append(u)
    if _unsupported:
        raise ConfigurationError(
            "Detcted config settings that are not compatible with this parser: %s."
            % _unsupported
        )

    _secret_cookiesigner = (options.get("secret"), options.get("cookie_signer"))
    if all(_secret_cookiesigner) or not any(_secret_cookiesigner):
        raise ConfigurationError(
            "One, and only one, of `redis.sessions.secret` and `redis.sessions.cookie_signer` must be provided."
        )

    # coerce bools
    for b in configs_bool:
        if b in options:
            options[b] = asbool(options[b])

    # coerce ints
    for i in configs_int:
        if i in options:
            options[i] = int(options[i])

    # allow "None" to be a value for some ints
    for i in configs_int_none:
        if i in options:
            if options[i] == "None":
                options[i] = None
            else:
                options[i] = int(options[i])
                if not options[i]:
                    options[i] = None

    # coerce float
    if "redis_socket_timeout" in options:
        options["redis_socket_timeout"] = float(options["redis_socket_timeout"])

    # check for settings conflict
    if "prefix" in options and "id_generator" in options:
        err = "cannot specify custom id_generator and a key prefix"
        raise ConfigurationError(err)

    # convenience setting for overriding key prefixes
    if "prefix" in options:
        prefix = options.pop("prefix")
        options["id_generator"] = partial(prefixed_id, prefix=prefix)

    return options


def persist(wrapped: Callable) -> Callable:
    """
    Decorator to persist in Redis all the data that needs to be persisted for
    this session and reset the expire time.
    This will mark the `_session_state.please_persist` as True, to be
    handled in a callback.
    To immediately persist a session, call `session.do_persist`.

    :param wrapped: a function to wrap with this decorator.
    :returns wrapped_refresh: a wrapped function.
    """

    def wrapped_persist(session: "RedisSession", *arg, **kw):
        result = wrapped(session, *arg, **kw)
        session._session_state.please_persist = True
        return result

    return wrapped_persist


def recookie(wrapped: Callable) -> Callable:
    """
    Decorator to mark a session as needing to recookie.
    This is necessary when setting a new max-age/etc

    :param wrapped: a function to wrap with this decorator.
    :returns wrapped_recookie: a wrapped function.
    """

    def wrapped_recookie(session: "RedisSession", *arg, **kw):
        result = wrapped(session, *arg, **kw)
        session._session_state.please_recookie = True
        return result

    return wrapped_recookie


def refresh(wrapped: Callable) -> Callable:
    """
    Decorator to reset the expire time for this session's key in Redis.
    This will mark the `_session_state.please_refresh` as True, to be
    handled in a callback.
    To immediately persist a session, call `session.do_refresh`.

    :param wrapped: a function to wrap with this decorator.
    :returns wrapped_refresh: a wrapped function.
    """

    def wrapped_refresh(session: "RedisSession", *arg, **kw):
        result = wrapped(session, *arg, **kw)
        session._session_state.please_refresh = True
        return result

    return wrapped_refresh


class SignedSerializerInterface(Protocol):
    def dumps(self, s: str) -> bytes: ...
    def loads(self, s: bytes) -> str: ...


# `SerializerInterface` ; DEPRECATED in 1.7; REMOVING in 1.8
SerializerInterface = SignedSerializerInterface


class _StringSerializer(SignedSerializerInterface):
    """
    A cheap serializer for compatibility with ``webob.cookies.SignedSerializer``.
    Our usage is only for encoding a signed session_id.

    By default, webob uses json loads/dumps.  As this library only uses strings
    for session, id, we can have a quick savings here.

    The webob interface dictates for ``SignedSerializer``
        https://github.com/Pylons/webob/blob/main/src/webob/cookies.py#L663

        An object with two methods: `loads`` and ``dumps``.  The ``loads`` method
        should accept bytes and return a Python object.  The ``dumps`` method
        should accept a Python object and return bytes.  A ``ValueError`` should
        be raised for malformed inputs.  Default: ``None`, which will use a
        derivation of :func:`json.dumps` and ``json.loads``.

    Additionally:
        SignedSerializer.dumps: expects Any and coverts to bytes;
           `SignedSerializer.dumps::self.serializer.dumps` expects Any and coverts to bytes;
                we only need to operate on strings, so can lock this down
        SignedSerializer.loads: expects Any and coverts to bytes;
           `SignedSerializer.dumps::self.serializer.dumps` expects Any and coverts to bytes;
                we only need to operate on strings, so can lock this down
    """

    def dumps(self, s: str) -> bytes:
        try:
            return s.encode("utf-8", "strict")
        except Exception as exc:
            raise InvalidSessionId_Serialization(exc)

    def loads(self, s: bytes) -> str:
        try:
            return str(s, "utf-8", "strict")
        except Exception as exc:
            raise InvalidSessionId_Deserialization(exc)


# `_NullSerializer` DEPRECATED in 1.7; REMOVING in 1.8
_NullSerializer = _StringSerializer



# -*- coding: utf-8 -*-
# stdlib
import pickle
from typing import Any
from typing import Dict
from typing import Optional

# pypi
from redis.exceptions import WatchError


# ==============================================================================


class DummySessionState(object):
    please_persist = None
    please_refresh = None


class DummySession(object):
    def __init__(
        self,
        session_id,
        redis,
        timeout=300,
        serialize=pickle.dumps,
    ):
        self.session_id = session_id
        self.redis = redis
        self.timeout = timeout
        self.serialize = serialize
        self.managed_dict = {}
        self.created = float()
        self._set_redis_ttl = True
        self._session_state = DummySessionState()

    def to_redis(self):
        data = {"m": self.managed_dict, "c": self.created}
        if self.timeout:
            data["t"] = self.timeout
        return self.serialize(data)


class DummyRedis(object):
    url: Optional[str]
    timeouts: Dict[str, int]
    _store: Dict[str, Any]
    opts: Dict[str, Any]

    def __init__(self, raise_watcherror=False, **kw):
        self.url = None
        self.timeouts = {}
        self._store = {}
        self.__dict__.update(kw)
        self._history = []
        self.pipeline = lambda: DummyPipeline(
            self._store, self, raise_watcherror
        )  # noqa: E501

    def _history_reset(self):
        # test method. fake. used for tests against the actual redis operations
        self._history = []

    @classmethod
    def from_url(cls, url, **opts):
        redis = DummyRedis()
        redis.url = url
        redis.opts = opts
        return redis

    def get(self, key):
        self._history.append(("get", key))
        return self._store.get(key)

    def set(self, key, value, debug=None):
        self._store[key] = value
        self._history.append(("set", key, value, debug))

    def setex(self, key, timeout, value, debug=None):
        # Redis is `key, value, timeout`
        # StrictRedis is `key, timeout, value`
        # this package uses StrictRedis
        self._store[key] = value
        self.timeouts[key] = timeout
        self._history.append(("setex", key, timeout, value, debug))

    def delete(self, *keys):
        for key in keys:
            del self._store[key]

    def exists(self, key):
        return key in self._store

    def expire(self, key, timeout):
        self.timeouts[key] = timeout
        self._history.append(("expire", key, timeout))

    def ttl(self, key):
        return self.timeouts.get(key)

    def keys(self):
        return self._store.keys()


class DummyPipeline(object):
    def __init__(self, store, redis_con, raise_watcherror=False):
        self._store = store
        self.raise_watcherror = raise_watcherror
        self._redis_con = redis_con
        self._history = []

    def __enter__(self):
        return self

    def __exit__(self, *arg, **kwarg):
        pass

    def multi(self):
        pass

    def set(self, key, value, debug=None):
        self._store[key] = value
        self._history.append(("set", key, debug))
        self._redis_con._history.append(("pipeline.set", key, value, debug))

    def get(self, key):
        self._history.append(("get", key))
        self._redis_con._history.append(("pipeline.get", key))
        return self._store.get(key)

    def expire(self, key, timeout):
        self._history.append(("expire", key, timeout))
        self._redis_con._history.append(("pipeline.expire", key, timeout))

    def setex(self, key, timeout, value, debug=None):
        self._store[key] = value
        self._history.append(("setex", key, timeout, debug))
        self._redis_con._history.append(
            ("pipeline.setex", key, timeout, value, debug)
        )  # noqa: E501

    def watch(self, key):
        if self.raise_watcherror:
            raise WatchError

    def execute(self):
        pass



# -*- coding: utf-8 -*-

# stdlib
import os
from typing import Any
from typing import Dict
from typing import Optional
from typing import Tuple
from typing import TYPE_CHECKING
import unittest

# pypi
from webob.cookies import SignedSerializer

# local
from pyramid_session_redis import _get_session_id_from_cookie
from pyramid_session_redis.exceptions import InvalidSessionId_Serialization
from pyramid_session_redis.util import _StringSerializer

# ==============================================================================

# export GENERATE_COOKIE_DATA=1
GENERATE_COOKIE_DATA = bool(int(os.getenv("GENERATE_COOKIE_DATA", "0")))

TYPE_COOKIES = Optional[Dict[str, str]]

# ----


class _ExpectedFailure_Setup(Exception):
    pass


class _FakeRequest(object):
    response: Optional["_FakeResponse"] = None

    @property
    def cookies(self) -> TYPE_COOKIES:
        """proxy against response.set_cookie to loosely mimic interace"""
        if self.response:
            return self.response.cookies
        return None


class _FakeResponse(object):
    req: Optional[_FakeRequest] = None
    cookies: TYPE_COOKIES = None

    def __init__(self, req: Optional[_FakeRequest] = None) -> None:
        self.cookies: TYPE_COOKIES = {}
        self.req = req
        if self.req:
            self.req.response = self

    def set_cookie(self, name: str, value: str) -> None:
        # webob.set_cookie expects str
        assert isinstance(value, str)
        if self.cookies is not None:
            self.cookies[name] = value


class Test_GetSessionIdFromCookie(unittest.TestCase):
    """
    In production, we should NEVER use a _StringSerializer like this.
    Instead, we should always use a SignedSerializer.

    The _StringSerializer is only used in this context to test the function
    `_get_session_id_from_cookie`
    """

    def _makeOne(
        self,
        cookie_name: str,
        cookie_value: str,
    ) -> Tuple[_FakeRequest, _FakeResponse, _StringSerializer]:
        serializer = _StringSerializer()
        if isinstance(cookie_value, str):
            _cookie_signed = serializer.dumps(cookie_value)  # bytes
            _cookie_value = _cookie_signed.decode()  # set_cookie wants str
        else:
            self.assertRaises(
                InvalidSessionId_Serialization, serializer.dumps, cookie_value
            )
            raise _ExpectedFailure_Setup()
        req = _FakeRequest()
        res = _FakeResponse(req)
        res.set_cookie(cookie_name, _cookie_value)
        if TYPE_CHECKING:
            assert res.cookies is not None
        self.assertIn(cookie_name, res.cookies)
        _value_encoded = res.cookies[cookie_name]
        self.assertEqual(type(_value_encoded), str)
        return req, res, serializer

    def _test_setup(
        self,
        cookie_name: str,
        value_in: Any,  # raw value in
        value_encoded: Any,  # encoded into Response; must be bytes
        value_decoded: str,  # decoded from request
    ) -> None:
        req, res, serializer = self._makeOne(cookie_name, value_in)
        if TYPE_CHECKING:
            assert res.cookies is not None
        _decoded = _get_session_id_from_cookie(req, cookie_name, serializer)
        self.assertIn(cookie_name, res.cookies)
        self.assertEqual(res.cookies[cookie_name], value_encoded)
        self.assertEqual(value_decoded, _decoded)
        self.assertEqual(type(_decoded), str)

    def test_string(self):
        cookie_name = "test_string"
        cookie_value = "string"
        value_encoded = "string"
        value_decoded = "string"
        self._test_setup(cookie_name, cookie_value, value_encoded, value_decoded)

    def test_int(self):
        cookie_name = "test_int"
        cookie_value = 1
        # this should fail during setup
        value_encoded = None
        value_decoded = None
        self.assertRaises(
            _ExpectedFailure_Setup,
            self._test_setup,
            cookie_name,
            cookie_value,
            value_encoded,
            value_decoded,
        )

    def test_bytes(self):
        cookie_name = "test_bytes"
        cookie_value = b"bytes"
        # this should fail during setup
        value_encoded = None
        value_decoded = None
        self.assertRaises(
            _ExpectedFailure_Setup,
            self._test_setup,
            cookie_name,
            cookie_value,
            value_encoded,
            value_decoded,
        )


class Test_CookieSigner_StringSerializer(unittest.TestCase):
    """
    to generate data for tests::

       export GENERATE_COOKIE_DATA=1
    """

    def _makeOne(
        self,
        cookie_name: str,
        cookie_value: str,
    ) -> Tuple[_FakeRequest, _FakeResponse, SignedSerializer]:
        # webob docs `SignedSerializer`:
        # An object with two methods: `loads`` and ``dumps``.  The ``loads`` method
        # should accept bytes and return a Python object.  The ``dumps`` method
        # should accept a Python object and return bytes.

        # SignedSerializer(secret, salt, hashalg="sha512", serializer=None)
        cookie_signer = SignedSerializer(
            "secret",
            "pyramid_session_redis.",
            "sha512",
            serializer=_StringSerializer(),
        )
        if isinstance(cookie_value, str):
            _cookie_signed = cookie_signer.dumps(cookie_value)  # bytes
            _cookie_value = _cookie_signed.decode()  # set_cookie( wants str
        else:
            self.assertRaises(
                InvalidSessionId_Serialization, cookie_signer.dumps, cookie_value
            )
            raise _ExpectedFailure_Setup()

        req = _FakeRequest()
        res = _FakeResponse(req)
        res.set_cookie(cookie_name, _cookie_value)
        if GENERATE_COOKIE_DATA:
            print("--------")
            print("%s.%s" % (self.__class__.__name__, self._testMethodName))
            print("  name:  ", cookie_name)
            print("  value: ", cookie_value)
            print("  signed:", _cookie_value)
        return req, res, cookie_signer

    def _test_setup(
        self,
        cookie_name: str,
        value_in: Any,
        value_expected: Any,
        value_signed: str,
    ) -> None:
        req, res, cookie_signer = self._makeOne(cookie_name, value_in)
        if TYPE_CHECKING:
            assert res.cookies is not None
        self.assertIn(cookie_name, res.cookies)
        self.assertEqual(res.cookies[cookie_name], value_signed)
        value_out = _get_session_id_from_cookie(req, cookie_name, cookie_signer)
        self.assertEqual(value_expected, value_out)

    def test_string(self):
        cookie_name = "test_string"
        cookie_value = "string"
        value_expected = "string"
        value_signed = "JW5R1E8zj18d1HHYi1v0bXsBrGgEPSG0BKQ4U1CjLkNbFM4WlTz1XzQEebEOv4k9F8RGuT3V_TizoDdRwz52sHN0cmluZw"
        self._test_setup(cookie_name, cookie_value, value_expected, value_signed)

    def test_bytes(self):
        cookie_name = "test_bytes"
        cookie_value = b"string"
        value_expected = None
        value_signed = None
        self.assertRaises(
            _ExpectedFailure_Setup,
            self._test_setup,
            cookie_name,
            cookie_value,
            value_expected,
            value_signed,
        )

    def test_int(self):
        cookie_name = "test_int"
        cookie_value = 1
        value_expected = None
        value_signed = None
        self.assertRaises(
            _ExpectedFailure_Setup,
            self._test_setup,
            cookie_name,
            cookie_value,
            value_expected,
            value_signed,
        )


class Test_CookieSigner_DefaultSerializer(unittest.TestCase):
    """
    to generate data for tests::

       export GENERATE_COOKIE_DATA=1
    """

    def _makeOne(
        self,
        cookie_name: str,
        cookie_value: str,
    ) -> Tuple[_FakeRequest, _FakeResponse, SignedSerializer]:
        # webob docs `SignedSerializer`:
        # An object with two methods: `loads`` and ``dumps``.  The ``loads`` method
        # should accept bytes and return a Python object.  The ``dumps`` method
        # should accept a Python object and return bytes.

        # SignedSerializer(secret, salt, hashalg="sha512", serializer=None)
        cookie_signer = SignedSerializer(
            "secret",
            "pyramid_session_redis.",
            "sha512",
        )
        if isinstance(cookie_value, bytes):
            self.assertRaises(TypeError, cookie_signer.dumps, cookie_value)
            raise _ExpectedFailure_Setup()
        else:
            _cookie_signed = cookie_signer.dumps(cookie_value)  # bytes
            _cookie_value = _cookie_signed.decode()  # .set_cookie wants str
        req = _FakeRequest()
        res = _FakeResponse(req)
        res.set_cookie(cookie_name, _cookie_value)
        if GENERATE_COOKIE_DATA:
            print("--------")
            print("%s.%s" % (self.__class__.__name__, self._testMethodName))
            print("  name:  ", cookie_name)
            print("  value: ", cookie_value)
            print("  signed:", _cookie_value)
        return req, res, cookie_signer

    def _test_setup(
        self,
        cookie_name: str,
        value_in: Any,
        value_expected: Any,
        value_signed: str,
    ) -> None:
        req, res, cookie_signer = self._makeOne(cookie_name, value_in)
        if TYPE_CHECKING:
            assert res.cookies is not None
        self.assertIn(cookie_name, res.cookies)
        self.assertEqual(res.cookies[cookie_name], value_signed)
        value_out = _get_session_id_from_cookie(req, cookie_name, cookie_signer)
        self.assertEqual(value_expected, value_out)

    def test_string(self):
        cookie_name = "test_string"
        cookie_value = "string"
        value_expected = "string"
        value_signed = "r3pVu9XGVFfz1MZQYdVT9kWVrb94mZT1TdL8HNYnFVf5cDcXPaL4ULuQ_GZ7hNAZNQCwfSRSGuffr6eQTgoHBSJzdHJpbmci"
        self._test_setup(cookie_name, cookie_value, value_expected, value_signed)

    def test_bytes(self):
        cookie_name = "test_bytes"
        cookie_value = b"bytes"
        value_expected = None
        value_signed = None
        self.assertRaises(
            _ExpectedFailure_Setup,
            self._test_setup,
            cookie_name,
            cookie_value,
            value_expected,
            value_signed,
        )

    def test_int(self):
        cookie_name = "test_int"
        cookie_value = 1
        value_expected = 1
        value_signed = "5aK2BseK-h0dYZa18Pxv8PVdLrKJwlmYNyh2Ck_-febyiABrjgFx1bNrwf128CL0I7Ulpw3f9FpOwlU7sMe5xDE"
        self._test_setup(cookie_name, cookie_value, value_expected, value_signed)


class Test_CookieSigner_Invalids(unittest.TestCase):
    """
    let's make sure we can handle completely invalid data correctly

    to generate the data for theses tests, just grab a signed value from elsewhere and change a character.
    below, we change the last character of a valid payload from 'i' to 'J'
    """

    def test_webob_default(self):
        # cookie_name = "test_string"
        # mess up a value from something else
        value_signed = b"r3pVu9XGVFfz1MZQYdVT9kWVrb94mZT1TdL8HNYnFVf5cDcXPaL4ULuQ_GZ7hNAZNQCwfSRSGuffr6eQTgoHBSJzdHJpbmcJ"

        # now, let's try to read it...
        # webob docs `SignedSerializer`:
        # An object with two methods: `loads`` and ``dumps``.  The ``loads`` method
        # should accept bytes and return a Python object.  The ``dumps`` method
        # should accept a Python object and return bytes.

        # SignedSerializer(secret, salt, hashalg="sha512", serializer=None)
        cookie_signer = SignedSerializer(
            "secret",
            "pyramid_session_redis.",
            "sha512",
        )
        with self.assertRaises(ValueError) as ctx:
            _cookie_value = cookie_signer.loads(value_signed)  # noqa: F841
        self.assertEqual(ctx.exception.args[0], "Invalid signature")

    def test_our_default(self):
        # cookie_name = "test_string"
        # mess up a value from something else
        value_signed = b"r3pVu9XGVFfz1MZQYdVT9kWVrb94mZT1TdL8HNYnFVf5cDcXPaL4ULuQ_GZ7hNAZNQCwfSRSGuffr6eQTgoHBSJzdHJpbmcJ"

        # now, let's try to read it...
        # webob docs `SignedSerializer`:
        # An object with two methods: `loads`` and ``dumps``.  The ``loads`` method
        # should accept bytes and return a Python object.  The ``dumps`` method
        # should accept a Python object and return bytes.

        # SignedSerializer(secret, salt, hashalg="sha512", serializer=None)
        cookie_signer = SignedSerializer(
            "secret",
            "pyramid_session_redis.",
            "sha512",
            serializer=_StringSerializer(),
        )
        with self.assertRaises(ValueError) as ctx:
            _cookie_value = cookie_signer.loads(value_signed)  # noqa: F841
        self.assertEqual(ctx.exception.args[0], "Invalid signature")

    def test_crafted_malicious_1(self):
        """
        WRITE w/json;
        READ w/str
        """
        # cookie_name = "test_int"
        cookie_value = {"a": False, "b": {"c": 1, "d": "e"}}

        # by default this will use a JSON serialization
        # webob docs `SignedSerializer`:
        # An object with two methods: `loads`` and ``dumps``.  The ``loads`` method
        # should accept bytes and return a Python object.  The ``dumps`` method
        # should accept a Python object and return bytes.

        # SignedSerializer(secret, salt, hashalg="sha512", serializer=None)
        cookie_signer_write = SignedSerializer(
            "secret",
            "pyramid_session_redis.",
            "sha512",
        )
        _cookie_value = cookie_signer_write.dumps(cookie_value)

        # decoding with the same signer should create an int
        decoded_native = cookie_signer_write.loads(_cookie_value)
        self.assertIsInstance(decoded_native, dict)

        self.assertEqual(cookie_value, decoded_native)

        # but this only handles strings...
        # webob docs `SignedSerializer`:
        # An object with two methods: `loads`` and ``dumps``.  The ``loads`` method
        # should accept bytes and return a Python object.  The ``dumps`` method
        # should accept a Python object and return bytes.

        # SignedSerializer(secret, salt, hashalg="sha512", serializer=None)
        cookie_signer_read = SignedSerializer(
            "secret",
            "pyramid_session_redis.",
            "sha512",
            serializer=_StringSerializer(),
        )

        # decoding with the alternate signer should force this as a string
        decoded_trans = cookie_signer_read.loads(_cookie_value)
        self.assertIsInstance(decoded_trans, str)



# -*- coding: utf-8 -*-

# stdlib
import itertools
import pickle
import time
from typing import Callable
from typing import Optional
from typing import TYPE_CHECKING
from typing import Union
import unittest

# pypi
from pyramid import testing

# local
from pyramid_session_redis.exceptions import InvalidSession_PayloadLegacy
from pyramid_session_redis.exceptions import InvalidSession_PayloadTimeout
from pyramid_session_redis.session import RedisSession
from pyramid_session_redis.util import encode_session_payload
from pyramid_session_redis.util import int_time
from pyramid_session_redis.util import LazyCreateSession
from . import DummyRedis

if TYPE_CHECKING:
    from redis.client import Redis as RedisClient


# ==============================================================================


class TestRedisSession(unittest.TestCase):
    def _makeOne(
        self,
        redis: Union[DummyRedis, "RedisClient"],
        session_id: str,
        new: bool,
        func_new_session: Callable,
        serialize=pickle.dumps,
        deserialize=pickle.loads,
        detect_changes=True,
        set_redis_ttl=True,
        deserialized_fails_new=None,
        timeout_trigger=None,
        timeout=1200,
        python_expires=True,
        set_redis_ttl_readheavy=None,
    ) -> RedisSession:
        _set_redis_ttl_onexit = False
        if (timeout and set_redis_ttl) and (
            not timeout_trigger and not python_expires and not set_redis_ttl_readheavy
        ):
            _set_redis_ttl_onexit = True
        return RedisSession(
            redis=redis,
            session_id=session_id,
            new=new,
            new_session=func_new_session,
            serialize=serialize,
            deserialize=deserialize,
            detect_changes=detect_changes,
            set_redis_ttl=set_redis_ttl,
            set_redis_ttl_readheavy=set_redis_ttl_readheavy,
            _set_redis_ttl_onexit=_set_redis_ttl_onexit,
            deserialized_fails_new=deserialized_fails_new,
            timeout_trigger=timeout_trigger,
            timeout=timeout,
            python_expires=python_expires,
        )

    def _set_up_session_in_redis(
        self,
        redis: Union[DummyRedis, "RedisClient"],
        session_id: str,
        timeout: int = 0,
        session_dict: Optional[dict] = None,
        serialize=pickle.dumps,
    ) -> str:
        """
        Note: this will call `encode_session_payload` with the initial session
        data. On a typical test this will mean an extra initial call to
        `encode_session_payload``
        """
        if session_dict is None:
            session_dict = {}
        time_now = int_time()
        expires = time_now + timeout if timeout else 0
        payload = encode_session_payload(session_dict, time_now, timeout, expires)
        redis.set(session_id, serialize(payload))
        return session_id

    def _make_id_generator(self) -> Callable:
        ids = itertools.count(start=0, step=1)
        return lambda: str(next(ids))

    def _set_up_session_in_Redis_and_makeOne(
        self,
        session_id: Optional[str] = None,
        session_dict: Optional[dict] = None,
        new=True,
        timeout=300,
        detect_changes=True,
    ) -> RedisSession:
        redis = DummyRedis()
        id_generator = self._make_id_generator()
        if session_id is None:
            session_id = id_generator()
        self._set_up_session_in_redis(
            redis=redis,
            session_id=session_id,
            session_dict=session_dict,
            timeout=timeout,
        )
        func_new_session = lambda: self._set_up_session_in_redis(
            redis=redis,
            session_id=id_generator(),
            session_dict=session_dict,
            timeout=timeout,
        )
        return self._makeOne(
            redis=redis,
            session_id=session_id,
            new=new,
            func_new_session=func_new_session,
            detect_changes=detect_changes,
            timeout=timeout,
        )

    def test_init_new_session(self):
        session_id = "session_id"
        new = True
        timeout = 300
        inst = self._set_up_session_in_Redis_and_makeOne(
            session_id=session_id, new=new, timeout=timeout
        )
        self.assertEqual(inst.session_id, session_id)
        self.assertIs(inst.new, new)
        self.assertDictEqual(dict(inst), {})

    def test_init_existing_session(self):
        session_id = "session_id"
        session_dict = {"key": "value"}
        new = False
        timeout = 300
        inst = self._set_up_session_in_Redis_and_makeOne(
            session_id=session_id, session_dict=session_dict, new=new, timeout=timeout
        )
        self.assertEqual(inst.session_id, session_id)
        self.assertIs(inst.new, new)
        self.assertDictEqual(dict(inst), session_dict)

    def test_delitem(self):
        inst = self._set_up_session_in_Redis_and_makeOne()
        inst["key"] = "val"
        del inst["key"]
        inst.do_persist()
        session_dict_in_redis = inst.from_redis()["m"]
        self.assertNotIn("key", inst)
        self.assertNotIn("key", session_dict_in_redis)

    def test_setitem(self):
        inst = self._set_up_session_in_Redis_and_makeOne()
        inst["key"] = "val"
        inst.do_persist()
        session_dict_in_redis = inst.from_redis()["m"]
        self.assertIn("key", inst)
        self.assertIn("key", session_dict_in_redis)

    def test_getitem(self):
        inst = self._set_up_session_in_Redis_and_makeOne()
        inst["key"] = "val"
        inst.do_persist()
        session_dict_in_redis = inst.from_redis()["m"]
        self.assertEqual(inst["key"], session_dict_in_redis["key"])

    def test_contains(self):
        inst = self._set_up_session_in_Redis_and_makeOne()
        inst["key"] = "val"
        inst.do_persist()
        session_dict_in_redis = inst.from_redis()["m"]
        self.assertTrue("key" in inst)
        self.assertTrue("key" in session_dict_in_redis)

    def test_setdefault(self):
        inst = self._set_up_session_in_Redis_and_makeOne()
        result = inst.setdefault("key", "val")
        self.assertEqual(result, inst["key"])

    def test_keys(self):
        inst = self._set_up_session_in_Redis_and_makeOne()
        inst["key1"] = ""
        inst["key2"] = ""
        inst_keys = inst.keys()
        inst.do_persist()
        session_dict_in_redis = inst.from_redis()["m"]
        persisted_keys = session_dict_in_redis.keys()
        self.assertEqual(inst_keys, persisted_keys)

    def test_items(self):
        inst = self._set_up_session_in_Redis_and_makeOne()
        inst["a"] = 1
        inst["b"] = 2
        inst_items = inst.items()
        inst.do_persist()
        session_dict_in_redis = inst.from_redis()["m"]
        persisted_items = session_dict_in_redis.items()
        self.assertEqual(inst_items, persisted_items)

    def test_clear(self):
        inst = self._set_up_session_in_Redis_and_makeOne()
        inst["a"] = 1
        inst.clear()
        session_dict_in_redis = inst.from_redis()["m"]
        self.assertNotIn("a", inst)
        self.assertNotIn("a", session_dict_in_redis)

    def test_get(self):
        inst = self._set_up_session_in_Redis_and_makeOne()
        inst["key"] = "val"
        get_from_inst = inst.get("key")
        self.assertEqual(get_from_inst, "val")
        inst.do_persist()
        session_dict_in_redis = inst.from_redis()["m"]
        get_from_redis = session_dict_in_redis.get("key")
        self.assertEqual(get_from_inst, get_from_redis)

    def test_get_default(self):
        inst = self._set_up_session_in_Redis_and_makeOne()
        get_from_inst = inst.get("key", "val")
        self.assertEqual(get_from_inst, "val")
        session_dict_in_redis = inst.from_redis()["m"]
        get_from_redis = session_dict_in_redis.get("key", "val")
        self.assertEqual(get_from_inst, get_from_redis)

    def test_pop(self):
        inst = self._set_up_session_in_Redis_and_makeOne()
        inst["key"] = "val"
        popped = inst.pop("key")
        self.assertEqual(popped, "val")
        session_dict_in_redis = inst.from_redis()["m"]
        self.assertNotIn("key", session_dict_in_redis)

    def test_pop_default(self):
        inst = self._set_up_session_in_Redis_and_makeOne()
        popped = inst.pop("key", "val")
        self.assertEqual(popped, "val")

    def test_update(self):
        inst = self._set_up_session_in_Redis_and_makeOne()
        inst["a"] = 1
        to_be_updated = {"a": "overriden", "b": 2}
        inst.update(to_be_updated)
        self.assertEqual(inst["a"], "overriden")
        self.assertEqual(inst["b"], 2)
        inst.do_persist()
        session_dict_in_redis = inst.from_redis()["m"]
        self.assertEqual(session_dict_in_redis["a"], "overriden")
        self.assertEqual(session_dict_in_redis["b"], 2)

    def test_iter(self):
        inst = self._set_up_session_in_Redis_and_makeOne()
        keys = ["a", "b", "c"]
        for k in keys:
            inst[k] = k
        itered = list(inst.__iter__())
        itered.sort()
        self.assertEqual(keys, itered)

    def test_has_key(self):
        inst = self._set_up_session_in_Redis_and_makeOne()
        inst["actual_key"] = ""
        self.assertIn("actual_key", inst)
        self.assertNotIn("not_a_key", inst)

    def test_values(self):
        inst = self._set_up_session_in_Redis_and_makeOne()
        inst["a"] = 1
        inst["b"] = 2
        expected_values = [1, 2]
        actual_values = sorted(inst.values())
        self.assertEqual(actual_values, expected_values)

    def test_itervalues(self):
        inst = self._set_up_session_in_Redis_and_makeOne()
        inst["a"] = 1
        inst["b"] = 2
        itered = list(inst.itervalues())
        itered.sort()
        expected = [1, 2]
        self.assertEqual(expected, itered)

    def test_iteritems(self):
        inst = self._set_up_session_in_Redis_and_makeOne()
        inst["a"] = 1
        inst["b"] = 2
        itered = list(inst.iteritems())
        itered.sort()
        expected = [("a", 1), ("b", 2)]
        self.assertEqual(expected, itered)

    def test_iterkeys(self):
        inst = self._set_up_session_in_Redis_and_makeOne()
        inst["a"] = 1
        inst["b"] = 2
        itered = list(inst.iterkeys())
        itered.sort()
        expected = ["a", "b"]
        self.assertEqual(expected, itered)

    def test_popitem(self):
        inst = self._set_up_session_in_Redis_and_makeOne()
        inst["a"] = 1
        inst["b"] = 2
        popped = inst.popitem()
        options = [("a", 1), ("b", 2)]
        self.assertIn(popped, options)
        session_dict_in_redis = inst.from_redis()["m"]
        self.assertNotIn(popped, session_dict_in_redis)

    def test_IDict_instance_conforms(self):
        from pyramid.interfaces import IDict
        from zope.interface.verify import verifyObject

        inst = self._set_up_session_in_Redis_and_makeOne()
        verifyObject(IDict, inst)

    def test_created(self):
        inst = self._set_up_session_in_Redis_and_makeOne()
        created = inst.from_redis()["c"]
        self.assertEqual(inst.created, created)

    def test_timeout(self):
        inst = self._set_up_session_in_Redis_and_makeOne()
        timeout = inst.from_redis()["t"]
        self.assertEqual(inst.timeout, timeout)

    def test_invalidate(self):
        inst = self._set_up_session_in_Redis_and_makeOne()
        first_session_id = inst.session_id
        inst.invalidate()
        self.assertNotIn(first_session_id, inst.redis._store)
        self.assertIs(inst._invalidated, True)

    def test_dict_multilevel(self):
        inst = self._set_up_session_in_Redis_and_makeOne(session_id="test1")
        inst["dict"] = {"foo": {"bar": 1}}
        inst.do_persist()
        get_from_inst = inst["dict"]["foo"]["bar"]
        self.assertEqual(get_from_inst, 1)
        session_dict_in_redis = inst.from_redis()["m"]
        get_from_redis = session_dict_in_redis["dict"]["foo"]["bar"]
        self.assertEqual(get_from_redis, 1)
        inst["dict"]["foo"]["bar"] = 2
        inst.do_persist()
        session_dict_in_redis2 = inst.from_redis()["m"]
        get_from_redis2 = session_dict_in_redis2["dict"]["foo"]["bar"]
        self.assertEqual(get_from_redis2, 2)

    def test_dict_multilevel_detect_changes_on(self):
        inst = self._set_up_session_in_Redis_and_makeOne(
            session_id="test1", detect_changes=True
        )
        # set a base dict and ensure it worked
        inst["dict"] = {"foo": {"bar": 1}}
        inst.do_persist()
        get_from_inst = inst["dict"]["foo"]["bar"]
        self.assertEqual(get_from_inst, 1)
        # grab the dict and edit it
        session_dict_in_redis = inst.from_redis()["m"]
        get_from_redis = session_dict_in_redis["dict"]["foo"]["bar"]
        self.assertEqual(get_from_redis, 1)
        inst["dict"]["foo"]["bar"] = 2
        # ensure the change was detected
        should_persist = inst._session_state.should_persist(inst)
        self.assertTrue(should_persist)

    def test_dict_multilevel_detect_changes_off(self):
        inst = self._set_up_session_in_Redis_and_makeOne(
            session_id="test1", detect_changes=False
        )
        # set a base dict and ensure it worked
        inst["dict"] = {"foo": {"bar": 1}}
        inst.do_persist()
        get_from_inst = inst["dict"]["foo"]["bar"]
        self.assertEqual(get_from_inst, 1)
        # grab the dict and edit it
        session_dict_in_redis = inst.from_redis()["m"]
        get_from_redis = session_dict_in_redis["dict"]["foo"]["bar"]
        self.assertEqual(get_from_redis, 1)
        inst["dict"]["foo"]["bar"] = 2
        # ensure the change was NOT detected
        should_persist = inst._session_state.should_persist(inst)
        self.assertFalse(should_persist)

    def test_new_session_after_invalidate(self):
        inst = self._set_up_session_in_Redis_and_makeOne()
        first_session_id = inst.session_id
        inst["key"] = "value"
        inst.invalidate()
        inst.ensure_id()  # ensure we have an id in redis, which creates a null payload
        second_session_id = inst.session_id
        self.assertSetEqual(set(inst.redis._store.keys()), {second_session_id})
        self.assertNotEqual(second_session_id, first_session_id)
        self.assertIs(bool(second_session_id), True)
        self.assertDictEqual(dict(inst), {})
        self.assertIs(inst.new, True)
        self.assertIs(inst._invalidated, False)

    def test_session_id_access_after_invalidate_creates_new_session(self):
        inst = self._set_up_session_in_Redis_and_makeOne()
        first_session_id = inst.session_id
        inst.invalidate()

        # 1.4.x+| session_id defaults to a LAZYCREATE
        self.assertIs(inst.session_id_safecheck, None)
        self.assertIs(inst.session_id, LazyCreateSession)

        second_session_id = inst.session_id
        self.assertNotEqual(second_session_id, first_session_id)
        self.assertIs(bool(second_session_id), True)

    def test_managed_dict_access_after_invalidate_creates_new_session(self):
        inst = self._set_up_session_in_Redis_and_makeOne()
        first_session_id = inst.session_id
        inst.invalidate()
        inst.managed_dict  # access

        # 1.4.x+| session_id defaults to a LAZYCREATE
        # 1.4.x+| session_id is only created via ensure_id()
        self.assertIs(inst.session_id_safecheck, None)
        self.assertIs(inst.session_id, LazyCreateSession)
        inst.ensure_id()

        # ORIGINALLY
        # .session_id attribute access also creates a new session after
        # invalidate, so just asserting .session_id is not enough to prove that
        # a new session was created after .managed_dict access. Here we note
        # down the session_ids in Redis right after .managed_dict access for an
        # additional check.
        session_ids_in_redis = inst.redis._store.keys()
        second_session_id = inst.session_id
        self.assertSetEqual(set(session_ids_in_redis), {second_session_id})
        self.assertNotEqual(second_session_id, first_session_id)
        self.assertIs(bool(second_session_id), True)

    def test_created_access_after_invalidate_creates_new_session(self):
        inst = self._set_up_session_in_Redis_and_makeOne()
        first_session_id = inst.session_id
        inst.invalidate()
        inst.created  # access

        # 1.4.x+| session_id defaults to a LAZYCREATE
        # 1.4.x+| session_id is only created via ensure_id()
        self.assertIs(inst.session_id_safecheck, None)
        self.assertIs(inst.session_id, LazyCreateSession)
        inst.ensure_id()

        # ORIGINALLY
        # .session_id attribute access also creates a new session after
        # invalidate, so just asserting .session_id was not enough to prove that
        # a new session was created after .created access. Here we noted down
        # the session_ids in Redis right after .created access for an
        # additional check.
        session_ids_in_redis = inst.redis._store.keys()
        second_session_id = inst.session_id
        self.assertSetEqual(set(session_ids_in_redis), {second_session_id})
        self.assertNotEqual(second_session_id, first_session_id)
        self.assertIs(bool(second_session_id), True)

    def test_timeout_access_after_invalidate_creates_new_session(self):
        inst = self._set_up_session_in_Redis_and_makeOne()
        first_session_id = inst.session_id
        inst.invalidate()
        inst.timeout  # access

        # 1.4.x+| session_id defaults to a LAZYCREATE
        # 1.4.x+| session_id is only created via ensure_id()
        self.assertIs(inst.session_id_safecheck, None)
        self.assertIs(inst.session_id, LazyCreateSession)
        inst.ensure_id()

        # ORIGINALLY:
        # .session_id attribute access also creates a new session after
        # invalidate, so just asserting .session_id is not enough to prove that
        # a new session was created after .timeout access. Here we note down
        # the session_ids in Redis right after .timeout access for an
        # additional check.
        session_ids_in_redis = inst.redis._store.keys()
        second_session_id = inst.session_id
        self.assertSetEqual(set(session_ids_in_redis), {second_session_id})
        self.assertNotEqual(second_session_id, first_session_id)
        self.assertIs(bool(second_session_id), True)

    def test_new_attribute_access_after_invalidate_creates_new_session(self):
        inst = self._set_up_session_in_Redis_and_makeOne()
        first_session_id = inst.session_id
        inst.invalidate()
        inst.new  # access

        # 1.4.x+| session_id defaults to a LAZYCREATE
        # 1.4.x+| session_id is only created via ensure_id()
        self.assertIs(inst.session_id_safecheck, None)
        self.assertIs(inst.session_id, LazyCreateSession)
        inst.ensure_id()

        # ORIGINALLY
        # .session_id attribute access also creates a new session after
        # invalidate, so just asserting .session_id is not enough to prove that
        # a new session was created after .created access. Here we note down
        # session_ids in Redis right after .new access for an additional check.
        session_ids_in_redis = inst.redis._store.keys()
        second_session_id = inst.session_id
        self.assertSetEqual(set(session_ids_in_redis), {second_session_id})
        self.assertNotEqual(second_session_id, first_session_id)
        self.assertIs(bool(second_session_id), True)

    def test_repeated_invalidates_without_new_session_created_in_between(self):
        inst = self._set_up_session_in_Redis_and_makeOne()
        first_session_id = inst.session_id
        inst.invalidate()
        inst.invalidate()
        self.assertNotIn(first_session_id, inst.redis._store)
        self.assertIs(inst._invalidated, True)
        second_session_id = inst.session_id
        self.assertNotEqual(second_session_id, first_session_id)
        self.assertIs(bool(second_session_id), True)

    def test_invalidate_new_session_invalidate(self):
        inst = self._set_up_session_in_Redis_and_makeOne()
        inst.invalidate()
        second_session_id = inst.session_id
        inst.invalidate()
        session_ids_in_redis = inst.redis._store.keys()
        self.assertNotIn(second_session_id, session_ids_in_redis)
        self.assertIs(inst._invalidated, True)

    def test_invalidate_new_session_invalidate_new_session(self):
        inst = self._set_up_session_in_Redis_and_makeOne()
        inst.ensure_id()  # ensure we have an id in redis, which creates a null payload

        inst.invalidate()
        inst.ensure_id()  # ensure we have an id in redis, which creates a null payload
        second_session_id = inst.session_id

        inst.invalidate()
        inst.ensure_id()  # ensure we have an id in redis, which creates a null payload
        third_session_id = inst.session_id

        session_ids_in_redis = inst.redis._store.keys()
        self.assertSetEqual(set(session_ids_in_redis), {third_session_id})
        self.assertNotEqual(third_session_id, second_session_id)
        self.assertIs(bool(third_session_id), True)
        self.assertDictEqual(dict(inst), {})
        self.assertIs(inst.new, True)
        self.assertIs(inst._invalidated, False)

    def test_mutablevalue_changed(self):
        inst = self._set_up_session_in_Redis_and_makeOne()
        inst["a"] = {"1": 1, "2": 2}
        tmp = inst["a"]
        tmp["3"] = 3
        inst.changed()
        inst.do_persist()
        session_dict_in_redis = inst.from_redis()["m"]
        self.assertEqual(session_dict_in_redis["a"], {"1": 1, "2": 2, "3": 3})

    def test_csrf_token(self):
        inst = self._set_up_session_in_Redis_and_makeOne()
        new_token = inst.new_csrf_token()
        got_token = inst.get_csrf_token()
        self.assertEqual(new_token, got_token)

    def test_get_new_csrf_token(self):
        inst = self._set_up_session_in_Redis_and_makeOne()
        self.assertNotIn("_csrft_", inst)
        token = inst.get_csrf_token()
        self.assertEqual(inst["_csrft_"], token)

    def test_flash(self):
        inst = self._set_up_session_in_Redis_and_makeOne()
        inst.flash("message")
        msgs = inst.peek_flash()
        self.assertIn("message", msgs)

    def test_flash_alternate_queue(self):
        inst = self._set_up_session_in_Redis_and_makeOne()
        inst.flash("message", "queue")
        default_queue = inst.peek_flash()
        self.assertNotIn("message", default_queue)
        other_queue = inst.peek_flash("queue")
        self.assertIn("message", other_queue)

    def test_pop_flash(self):
        inst = self._set_up_session_in_Redis_and_makeOne()
        inst.flash("message")
        popped = inst.pop_flash()
        self.assertIn("message", popped)
        msgs = inst.peek_flash()
        self.assertEqual(msgs, [])

    def test_ISession_instance_conforms(self):
        from pyramid.interfaces import ISession
        from zope.interface.verify import verifyObject

        inst = self._set_up_session_in_Redis_and_makeOne()
        verifyObject(ISession, inst)

    def _test_adjust_session_timeout(self, variant: Optional[str] = None):
        inst = self._set_up_session_in_Redis_and_makeOne(timeout=100)
        adjusted_timeout = 200
        if not variant:
            inst.adjust_session_timeout(adjusted_timeout)
        else:
            getattr(inst, variant)(adjusted_timeout)
        inst.do_persist()
        self.assertEqual(inst.timeout, adjusted_timeout)
        self.assertEqual(inst.from_redis()["t"], adjusted_timeout)

    def test_adjust_session_timeout(self):
        self._test_adjust_session_timeout()

    def test_adjust_session_timeout__legacy(self):
        self._test_adjust_session_timeout(variant="adjust_timeout_for_session")


class _TestRedisSessionNew_CORE(object):
    """
    Support functions for 1.2.x+ tests
    see
    """

    session_id = "session_id"

    def _makeOne(
        self,
        redis: Union[DummyRedis, "RedisClient"],
        session_id: str,
        new: bool,
        func_new_session: Callable,
        serialize=pickle.dumps,
        deserialize=pickle.loads,
        detect_changes=True,
        set_redis_ttl=True,
        deserialized_fails_new: Optional[bool] = None,
        timeout_trigger: Optional[int] = None,
        timeout=1200,
        python_expires=True,
        set_redis_ttl_readheavy: Optional[bool] = None,
    ) -> RedisSession:
        _set_redis_ttl_onexit = False
        if (timeout and set_redis_ttl) and (
            not timeout_trigger and not python_expires and not set_redis_ttl_readheavy
        ):
            _set_redis_ttl_onexit = True

        return RedisSession(
            redis=redis,
            session_id=session_id,
            new=new,
            new_session=func_new_session,
            serialize=serialize,
            deserialize=deserialize,
            detect_changes=detect_changes,
            set_redis_ttl=set_redis_ttl,
            set_redis_ttl_readheavy=set_redis_ttl_readheavy,
            _set_redis_ttl_onexit=_set_redis_ttl_onexit,
            deserialized_fails_new=deserialized_fails_new,
            timeout_trigger=timeout_trigger,
            timeout=timeout,
            python_expires=python_expires,
        )

    def _set_up_session_in_redis(
        self,
        redis: Union[DummyRedis, "RedisClient"],
        session_id: str,
        timeout: int = 0,
        session_dict: Optional[dict] = None,
        serialize=pickle.dumps,
        session_version: Optional[int] = None,
        expires: int = 0,
        python_expires=True,
        reset_history=True,
    ) -> str:
        """
        Note: this will call `encode_session_payload` with the initial session
        data. On a typical test this will mean an extra initial call to
        `encode_session_payload`.
        the `expires` kwarg is to fake an expiry date.
        """
        if session_dict is None:
            session_dict = {}
        time_now = int_time()
        payload = encode_session_payload(
            session_dict, time_now, timeout, expires, python_expires=python_expires
        )
        if session_version is not None:
            payload["v"] = session_version
        # code an expires if needed
        if expires:
            # `encode_session_payload` will update the expires if we passed it,
            # so we must insert an expired value manually
            if python_expires:
                payload["x"] = expires
        elif not expires:
            if python_expires:
                expires = time_now + timeout
                payload["x"] = expires
        redis.set(session_id, serialize(payload))
        if reset_history:
            redis._history_reset()
        return session_id

    def _make_id_generator(self) -> Callable:
        ids = itertools.count(start=0, step=1)
        return lambda: str(next(ids))

    def _set_up_session_in_Redis_and_makeOne(
        self,
        session_id: Optional[str] = None,
        session_dict: Optional[dict] = None,
        new: bool = True,
        timeout: int = 60,
        timeout_trigger: int = 30,
        detect_changes=True,
        set_redis_ttl=True,
        session_version: Optional[int] = None,
        expires: int = 0,
        python_expires=True,
        set_redis_ttl_readheavy: Optional[bool] = None,
    ) -> RedisSession:
        redis = DummyRedis()
        id_generator = self._make_id_generator()
        if session_id is None:
            session_id = id_generator()
        self._set_up_session_in_redis(
            redis=redis,
            session_id=session_id,
            session_dict=session_dict,
            timeout=timeout,
            session_version=session_version,
            expires=expires,
            python_expires=python_expires,
        )
        func_new_session = lambda: self._set_up_session_in_redis(
            redis=redis,
            session_id=id_generator(),
            session_dict=session_dict,
            timeout=timeout,
            python_expires=python_expires,
        )
        return self._makeOne(
            redis=redis,
            session_id=session_id,
            new=new,
            func_new_session=func_new_session,
            detect_changes=detect_changes,
            timeout=timeout,
            set_redis_ttl=set_redis_ttl,
            set_redis_ttl_readheavy=set_redis_ttl_readheavy,
            python_expires=python_expires,
        )

    def _deserialize_session(
        self,
        session: RedisSession,
        deserialize=pickle.loads,
    ) -> dict:
        _session_id = session.session_id
        _session_data = session.redis._store[_session_id]
        _session_deserialized = deserialize(_session_data)
        return _session_deserialized


class _TestRedisSessionNew__MIXIN_A(object):

    # TODO: typing/protocol for these expected mixin methods
    _set_up_session_in_Redis_and_makeOne: Callable
    _make_id_generator: Callable
    _makeOne: Callable
    _set_up_session_in_redis: Callable
    assertEqual: Callable
    assertIs: Callable
    assertDictEqual: Callable

    # class
    PYTHON_EXPIRES: Optional[bool] = None
    set_redis_ttl: Optional[bool] = None
    set_redis_ttl_readheavy: Optional[bool] = None
    session_id: str = "session_id"
    timeout: Optional[int] = 3
    timeout_trigger: Optional[int] = 6
    adjusted_timeout: Optional[int] = 6

    def _session_new(self) -> RedisSession:
        session = self._set_up_session_in_Redis_and_makeOne(
            session_id=self.session_id,
            new=True,
            timeout=self.timeout,
            set_redis_ttl=self.set_redis_ttl,
            set_redis_ttl_readheavy=self.set_redis_ttl_readheavy,
            python_expires=self.PYTHON_EXPIRES,
        )
        _request = testing.DummyRequest()
        session._deferred_callback(_request)  # trigger the real session's set/setex
        self.assertEqual(session.session_id, self.session_id)
        self.assertIs(session.new, True)
        self.assertDictEqual(dict(session), {})
        return session

    def _factory_new_session(self, session: RedisSession) -> Callable:
        id_generator = self._make_id_generator()
        func_new_session = lambda: self._set_up_session_in_redis(
            redis=session.redis,
            session_id=id_generator(),
            session_dict={},
            timeout=self.timeout,
            python_expires=self.PYTHON_EXPIRES,
            set_redis_ttl=self.set_redis_ttl,
            set_redis_ttl_readheavy=self.set_redis_ttl_readheavy,
        )
        return func_new_session

    def _session_get(
        self, session: RedisSession, func_new_session: Callable
    ) -> RedisSession:
        session2 = self._makeOne(
            session.redis,
            self.session_id,
            True,
            func_new_session,
            set_redis_ttl=self.set_redis_ttl,
            set_redis_ttl_readheavy=self.set_redis_ttl_readheavy,
            timeout=self.timeout,
            python_expires=self.PYTHON_EXPIRES,
        )
        return session2


class TestRedisSessionNew(
    unittest.TestCase,
    _TestRedisSessionNew_CORE,
):
    """these are 1.2x+ tests"""

    def test_init_new_session_notimeout(self):
        new = True
        timeout = 0
        set_redis_ttl = True
        session = self._set_up_session_in_Redis_and_makeOne(
            session_id=self.session_id,
            new=new,
            timeout=timeout,
            set_redis_ttl=set_redis_ttl,
        )
        session.do_persist()  # trigger the real session's set/setex
        self.assertEqual(session.session_id, self.session_id)
        self.assertIs(session.new, new)
        self.assertDictEqual(dict(session), {})

        self.assertEqual(session.timeout, None)

        _deserialized = self._deserialize_session(session)
        self.assertNotIn("t", _deserialized)

        # get, set
        self.assertEqual(len(session.redis._history), 2)
        _redis_op = session.redis._history[1]
        self.assertEqual(_redis_op[0], "set")

        # clear the history, `do_refresh` should do nothing (timeout=0)
        session.redis._history_reset()
        session.do_refresh()  # trigger the real session's set/setex
        self.assertEqual(
            len(session.redis._history), 0
        )  # we shouldn't have any timeout at all

        # clear the history, `do_refresh+force_redis_ttl` ensures an "expire"
        session.redis._history_reset()
        session.do_refresh(force_redis_ttl=47)  # trigger the real session's set/setex
        self.assertEqual(len(session.redis._history), 1)
        _redis_op = session.redis._history[0]
        self.assertEqual(_redis_op[0], "expire")
        self.assertEqual(_redis_op[2], 47)

    def test_init_new_session_notimeout_lru(self):
        """
        check that a no timeout will trigger a SET if LRU enabled
        """
        new = True
        timeout = 0
        set_redis_ttl = False
        session = self._set_up_session_in_Redis_and_makeOne(
            session_id=self.session_id,
            new=new,
            timeout=timeout,
            set_redis_ttl=set_redis_ttl,
        )
        session.do_persist()  # trigger the real session's set/setex
        self.assertEqual(session.session_id, self.session_id)
        self.assertIs(session.new, new)
        self.assertDictEqual(dict(session), {})

        _deserialized = self._deserialize_session(session)
        self.assertNotIn("t", _deserialized)

        # get, set
        self.assertEqual(len(session.redis._history), 2)
        _redis_op = session.redis._history[1]
        self.assertEqual(_redis_op[0], "set")

        # clear the history, `do_refresh` should do nothing (timeout=0)
        session.redis._history_reset()
        session.do_refresh()  # trigger the real session's set/setex
        self.assertEqual(
            len(session.redis._history), 0
        )  # we shouldn't have any timeout at all

        # clear the history, `do_refresh+force_redis_ttl` ensures an "expire"
        session.redis._history_reset()
        session.do_refresh(force_redis_ttl=47)  # trigger the real session's set/setex
        self.assertEqual(len(session.redis._history), 1)
        _redis_op = session.redis._history[0]
        self.assertEqual(_redis_op[0], "expire")
        self.assertEqual(_redis_op[2], 47)

    def test_init_new_session_timeout(self):
        """
        check that a timeout will trigger a SETEX
        """
        new = True
        timeout = 60
        set_redis_ttl = True
        session = self._set_up_session_in_Redis_and_makeOne(
            session_id=self.session_id,
            new=new,
            timeout=timeout,
            set_redis_ttl=set_redis_ttl,
        )
        session.do_persist()  # trigger the real session's set/setex
        self.assertEqual(session.session_id, self.session_id)
        self.assertIs(session.new, new)
        self.assertDictEqual(dict(session), {})

        _deserialized = self._deserialize_session(session)
        self.assertIn("t", _deserialized)
        # get, set
        self.assertEqual(len(session.redis._history), 2)
        _redis_op = session.redis._history[1]
        self.assertEqual(_redis_op[0], "setex")

        # clear the history, `do_refresh` should issue an "expire" (timeout=60)
        session.redis._history_reset()
        session.do_refresh()  # trigger the real session's set/setex
        # we shouldn't have set a timeout
        self.assertEqual(len(session.redis._history), 1)
        _redis_op = session.redis._history[0]
        self.assertEqual(_redis_op[0], "expire")

        # clear the history, `do_refresh+force_redis_ttl` ensures an "expire"
        session.redis._history_reset()
        session.do_refresh(force_redis_ttl=47)  # trigger the real session's set/setex
        self.assertEqual(len(session.redis._history), 1)
        _redis_op = session.redis._history[0]
        self.assertEqual(_redis_op[0], "expire")
        self.assertEqual(_redis_op[2], 47)

    def test_init_new_session_timeout_lru(self):
        """
        check that a timeout will trigger a SET if LRU enabled
        """
        new = True
        timeout = 60
        set_redis_ttl = False
        session = self._set_up_session_in_Redis_and_makeOne(
            session_id=self.session_id,
            new=new,
            timeout=timeout,
            set_redis_ttl=set_redis_ttl,
        )
        session.do_persist()  # trigger the real session's set/setex
        self.assertEqual(session.session_id, self.session_id)
        self.assertIs(session.new, new)
        self.assertDictEqual(dict(session), {})
        _deserialized = self._deserialize_session(session)
        self.assertIn("t", _deserialized)
        self.assertEqual(len(session.redis._history), 2)
        _redis_op = session.redis._history[1]
        self.assertEqual(_redis_op[0], "set")

        # clear the history, `do_refresh` should do nothing (timeout=60, set_redis_ttl=False)
        session.redis._history_reset()
        session.do_refresh()  # trigger the real session's set/setex
        self.assertEqual(len(session.redis._history), 0)

        # clear the history, `do_refresh` should issue an "expire" (force_redis_ttl=60)
        session.redis._history_reset()
        session.do_refresh(force_redis_ttl=47)  # trigger the real session's set/setex
        self.assertEqual(len(session.redis._history), 1)
        _redis_op = session.redis._history[0]
        self.assertEqual(_redis_op[0], "expire")
        self.assertEqual(_redis_op[2], 47)

    def test_session_invalid_legacy(self):
        """
        check that ``exceptions.InvalidSession_PayloadLegacy`` is raised if a previous version is detected
        """
        new = True
        timeout = 60
        set_redis_ttl = False
        session_version = -1
        with self.assertRaises(InvalidSession_PayloadLegacy):
            session = self._set_up_session_in_Redis_and_makeOne(  # noqa: F841
                session_id=self.session_id,
                new=new,
                timeout=timeout,
                set_redis_ttl=set_redis_ttl,
                session_version=session_version,
            )
        return

    def test_session_invalid_expires(self):
        """
        check that ``exceptions.InvalidSession_PayloadTimeout`` is raised if we timed out
        """
        new = True
        timeout = 60
        set_redis_ttl = False
        expires = int_time() - timeout - 1
        with self.assertRaises(InvalidSession_PayloadTimeout):
            session = self._set_up_session_in_Redis_and_makeOne(  # noqa: F841
                session_id=self.session_id,
                new=new,
                timeout=timeout,
                set_redis_ttl=set_redis_ttl,
                expires=expires,
            )


class TestRedisSessionNew_TimeoutAdjustments_A(
    _TestRedisSessionNew__MIXIN_A,
    _TestRedisSessionNew_CORE,
    unittest.TestCase,
):
    """these are 1.4x+ tests"""

    PYTHON_EXPIRES = True
    set_redis_ttl = True
    timeout = 6
    timeout_trigger = 3
    adjusted_timeout = 86400  # 31536000

    def test_adjust_timeout(self):
        """
        check that a timeout will trigger a SETEX

        python -munittest pyramid_session_redis.tests.test_session.TestRedisSessionNew_TimeoutAdjustments_A.test_adjust_timeout
        """
        self._test_adjust_timeout()

    def test_adjust_timeout__legacy(self):
        self._test_adjust_timeout(variant="adjust_timeout_for_session")

    def _test_adjust_timeout(self, variant: Optional[str] = None):
        session = self._session_new()
        serialized_1 = (
            session.from_redis()
        )  # this executes a second get, the session doesn't save the raw value

        timeout_1 = serialized_1["t"]
        self.assertEqual(timeout_1, self.timeout)
        timestamp_created = serialized_1["c"]
        timestamp_expiry_initial = serialized_1["x"]  # noqa: F841
        if not variant:
            session.adjust_session_timeout(self.adjusted_timeout)
        else:
            getattr(session, variant)(self.adjusted_timeout)
        _request = testing.DummyRequest()
        session._deferred_callback(_request)  # trigger the real session's set/setex
        self.assertEqual(len(session.redis._history), 3)
        self.assertEqual(session.redis._history[0][0], "get")
        self.assertEqual(session.redis._history[1][0], "get")
        self.assertEqual(session.redis._history[2][0], "setex")

        serialized_2 = (
            session.from_redis()
        )  # this executes a second get, the session doesn't save the raw value
        self.assertEqual(len(session.redis._history), 4)
        self.assertEqual(session.redis._history[3][0], "get")

        timestamp_expiry_new = serialized_2["x"]
        timeout_2 = serialized_2["t"]
        self.assertEqual(timeout_2, self.adjusted_timeout)

        timestamp_expiry_expected = timestamp_created + self.adjusted_timeout
        self.assertEqual(timestamp_expiry_new, timestamp_expiry_expected)

        # okay now check that redis got the correct commands
        # get, [get], setex, [get] || [] are our test commands, not part of the flow
        self.assertEqual(len(session.redis._history), 4)


class TestRedisSessionNew_TimeoutAdjustments_B(
    _TestRedisSessionNew__MIXIN_A,
    _TestRedisSessionNew_CORE,
    unittest.TestCase,
):
    """these are 1.4x+ tests"""

    PYTHON_EXPIRES = True
    set_redis_ttl = True
    timeout = 3
    timeout_trigger = 1
    adjusted_timeout = 4

    def test_timeout_trigger(self):
        self._test_timeout_trigger()

    def test_timeout_trigger__legacy(self):
        self._test_timeout_trigger(variant="adjust_timeout_for_session")

    def _test_timeout_trigger(self, variant: Optional[str] = None):
        """
        python -munittest pyramid_session_redis.tests.test_session.TestRedisSessionNew_TimeoutAdjustments_B.test_timeout_trigger
        """
        session = self._session_new()
        _request = testing.DummyRequest()

        # this should set:
        # session.managed_dict = {'FOO': '1'}
        session["FOO"] = "1"
        session._deferred_callback(_request)  # trigger the real session's set/setex

        serialized_1 = session.from_redis()
        timeout_1 = serialized_1["t"]
        self.assertEqual(timeout_1, self.timeout)
        timestamp_created = serialized_1["c"]
        timestamp_expiry_initial = serialized_1["x"]  # noqa: F841
        if not variant:
            session.adjust_session_timeout(self.adjusted_timeout)
        else:
            getattr(session, variant)(self.adjusted_timeout)
        session._deferred_callback(_request)  # trigger the real session's set/setex

        serialized_2 = session.from_redis()
        timestamp_expiry_new = serialized_2["x"]
        timeout_2 = serialized_2["t"]
        self.assertEqual(timeout_2, self.adjusted_timeout)

        timestamp_expiry_expected = timestamp_created + self.adjusted_timeout
        self.assertEqual(timestamp_expiry_new, timestamp_expiry_expected)

        # okay now check that redis got the correct commands
        # 0 - GET
        # 1 - SETEX
        # 2 - GET
        # 3 - SETEX
        # 4 - GET
        self.assertEqual(len(session.redis._history), 5)
        self.assertEqual(session.redis._history[1][0], "setex")  # this is our foo
        self.assertEqual(session.redis._history[1][2], self.timeout)
        self.assertEqual(session.redis._history[3][0], "setex")
        self.assertEqual(session.redis._history[3][2], self.adjusted_timeout)

        # sleep post-trigger and pre-expire
        # this should create a new trigger, because we wake up within 1 second...
        time.sleep(3)

        func_new_session = self._factory_new_session(session)
        session2 = self._session_get(session, func_new_session)
        self.assertEqual(session2.managed_dict["FOO"], "1")
        session2._deferred_callback(_request)  # trigger the real session's set/setex

        # okay now check that redis got the correct commands
        # 0 - GET
        # 1 - SETEX
        # 2 - GET
        # 3 - SETEX
        # 4 - GET
        # 5 - GET
        # 6 - SETEX -- this is triggered by the _deferred_callback beine within the timeout
        self.assertEqual(len(session.redis._history), 7)

        # now check that the timeout is in the payload...
        session2_redis = session2.from_redis()
        self.assertEqual(
            session2_redis["x"], (session2.timestamp + self.adjusted_timeout)
        )

        sleeptime = session2_redis["x"] - int(time.time()) + 1
        print("sleeping for ", sleeptime)
        time.sleep(sleeptime)

        with self.assertRaises(InvalidSession_PayloadTimeout):
            session3 = self._makeOne(  # noqa: F841
                session.redis,
                self.session_id,
                True,  # new
                func_new_session,
                set_redis_ttl=self.set_redis_ttl,
                timeout_trigger=self.timeout_trigger,
                timeout=self.timeout,
                python_expires=self.PYTHON_EXPIRES,
            )

        # okay now check that redis got the correct commands
        # 0 - GET
        # 1 - SETEX
        # 2 - GET
        # 3 - SETEX
        # 4 - GET
        # 5 - GET
        # 6 - SETEX -- this is triggered by the _deferred_callback beine within the timeout
        # 7 - GET
        # 8 - GET
        self.assertEqual(len(session.redis._history), 9)


class TestRedisSessionNew_RedisTTL_Classic(
    _TestRedisSessionNew__MIXIN_A,
    _TestRedisSessionNew_CORE,
    unittest.TestCase,
):
    """these are 1.4x+ tests"""

    PYTHON_EXPIRES = False
    set_redis_ttl = True
    set_redis_ttl_readheavy = False
    timeout = 1
    timeout_trigger = None

    def test_refresh(self):
        """
        python -munittest pyramid_session_redis.tests.test_session.TestRedisSessionNew_RedisTTL_Classic.test_refresh
        """
        session = self._session_new()
        _request = testing.DummyRequest()

        # okay now check that redis got the correct commands
        # 0 - GET
        # no setex, because the session is empty
        self.assertEqual(len(session.redis._history), 1)
        self.assertEqual(session.redis._history[0][0], "get")  # this is our foo

        # but if we store something it would have been different...
        session["FOO"] = "1"
        session._deferred_callback(_request)  # trigger the real session's set/setex

        # 0 - GET
        # 1 - SETEX
        self.assertEqual(len(session.redis._history), 2)
        self.assertEqual(session.redis._history[1][0], "setex")  # this is our foo
        self.assertEqual(session.redis._history[1][2], self.timeout)

        func_new_session = self._factory_new_session(session)
        session2 = self._session_get(session, func_new_session)
        self.assertEqual(session2.managed_dict["FOO"], "1")
        session2._deferred_callback(_request)  # trigger the real session's expire

        # 0 - GET
        # 1 - SETEX
        # 2 - GET
        # 3 - EXPIRE
        self.assertEqual(len(session.redis._history), 4)
        self.assertEqual(session.redis._history[2][0], "get")
        self.assertEqual(session.redis._history[3][0], "expire")

        # will sleep for a moment...
        sleepy = self.timeout - 1
        print("will sleep for:", sleepy)
        time.sleep(sleepy)

        session3 = self._session_get(session, func_new_session)
        self.assertEqual(session3.managed_dict["FOO"], "1")
        session3._deferred_callback(
            _request
        )  # trigger the real session's set/setex (but don't)

        # 0 - GET
        # 1 - SETEX
        # 2 - GET
        # 3 - EXPIRE
        # 4 - GET
        # 5 - EXPIRE
        self.assertEqual(len(session.redis._history), 6)
        self.assertEqual(session.redis._history[4][0], "get")
        self.assertEqual(session.redis._history[5][0], "expire")


class TestRedisSessionNew_RedisTTL_ReadHeavy(
    _TestRedisSessionNew__MIXIN_A,
    _TestRedisSessionNew_CORE,
    unittest.TestCase,
):
    """these are 1.4x+ tests"""

    PYTHON_EXPIRES = False
    set_redis_ttl = True
    set_redis_ttl_readheavy = True
    timeout = 4

    def test_refresh(self):
        """
        python -munittest pyramid_session_redis.tests.test_session.TestRedisSessionNew_RedisTTL_ReadHeavy.test_refresh
        """
        session = self._session_new()
        _request = testing.DummyRequest()

        # okay now check that redis got the correct commands
        # 0 - pipline.GET
        # 1 - pipline.EXPIRE
        # no setex, because the session is empty
        self.assertEqual(len(session.redis._history), 2)

        # but if we store something it would have been different...
        session["FOO"] = "1"
        session._deferred_callback(_request)  # trigger the real session's set/setex

        # 0 - pipline.GET
        # 1 - pipline.EXPIRE
        # 2 - SETEX
        self.assertEqual(len(session.redis._history), 3)
        self.assertEqual(session.redis._history[0][0], "pipeline.get")
        self.assertEqual(session.redis._history[1][0], "pipeline.expire")
        self.assertEqual(session.redis._history[2][0], "setex")  # this is our foo
        self.assertEqual(session.redis._history[2][2], self.timeout)

        func_new_session = self._factory_new_session(session)
        session2 = self._session_get(session, func_new_session)
        self.assertEqual(session2.managed_dict["FOO"], "1")
        session2._deferred_callback(
            _request
        )  # trigger the real session's set/setex (but don't)

        # 0 - pipline.GET
        # 1 - pipline.EXPIRE
        # 2 - SETEX
        # 3 - pipline.GET
        # 4 - pipline.EXPIRE
        self.assertEqual(len(session.redis._history), 5)
        self.assertEqual(session.redis._history[3][0], "pipeline.get")
        self.assertEqual(session.redis._history[4][0], "pipeline.expire")

        # will sleep for a moment...
        sleepy = self.timeout - 1
        print("will sleep for:", sleepy)
        time.sleep(sleepy)

        session3 = self._session_get(session, func_new_session)
        self.assertEqual(session3.managed_dict["FOO"], "1")
        session3._deferred_callback(
            _request
        )  # trigger the real session's set/setex (but don't)

        # 0 - pipline.GET
        # 1 - pipline.EXPIRE
        # 2 - SETEX
        # 3 - pipline.GET
        # 4 - pipline.EXPIRE
        # 5 - pipline.GET
        # 6 - pipline.EXPIRE
        self.assertEqual(len(session.redis._history), 7)
        self.assertEqual(session.redis._history[5][0], "pipeline.get")
        self.assertEqual(session.redis._history[6][0], "pipeline.expire")



# stdlib
from typing import Dict
from typing import Union

# local
from pyramid_session_redis.util import SignedSerializerInterface

# ------------------------------------------------------------------------------


# dotted paths to dummy callables
_id_path = "tests.test_config.dummy_id_generator"
_client_path = "tests.test_config.dummy_client_callable"
_invalid_logger = "tests.test_config.dummy_invalid_logger"

TEST_PSR_CONFIG: Dict[str, Union[str, int]] = {
    "redis.sessions.secret": "supersecret",
    "redis.sessions.db": 9,
    "redis.sessions.serialize": "pickle.dumps",
    "redis.sessions.deserialize": "pickle.loads",
    "redis.sessions.id_generator": _id_path,
    "redis.sessions.client_callable": _client_path,
    "redis.sessions.func_invalid_logger": _invalid_logger,
}

LIVE_PSR_CONFIG = TEST_PSR_CONFIG.copy()
del LIVE_PSR_CONFIG["redis.sessions.id_generator"]
del LIVE_PSR_CONFIG["redis.sessions.client_callable"]


class CustomCookieSigner(SignedSerializerInterface):

    def dumps(self, s: str) -> bytes:
        return s.encode()

    def loads(self, s: bytes) -> str:
        return s.decode()



# -*- coding: utf-8 -*-

# stdlib
import unittest

# pypi
from pyramid import testing
from pyramid.exceptions import ConfigurationError
from pyramid.request import Request
from pyramid.threadlocal import get_current_request
from typing_extensions import Literal

# local
from pyramid_session_redis.exceptions import InvalidSession
from pyramid_session_redis.exceptions import (
    InvalidSession_DeserializationError,
)  # noqa: E501
from ._util import _client_path
from ._util import _id_path
from ._util import _invalid_logger
from ._util import CustomCookieSigner

# ==============================================================================


# used to ensure includeme can resolve a dotted path to an id generator
def dummy_id_generator() -> str:
    return "42"


# used to ensure includeme can resolve a dotted path to a redis client callable
def dummy_client_callable(request: Request, **opts) -> str:
    return "client"


def dummy_invalid_logger(request: Request, raised: Exception) -> Literal[True]:
    assert isinstance(raised, InvalidSession)
    return True


# ------------------------------------------------------------------------------


class Test_includeme_simple(unittest.TestCase):
    def setUp(self):
        request = testing.DummyRequest()
        self.config = testing.setUp(request=request)
        self.config.registry.settings = {
            "redis.sessions.secret": "supersecret",
            "redis.sessions.db": 9,
            "redis.sessions.serialize": "pickle.dumps",
            "redis.sessions.deserialize": "pickle.loads",
            "redis.sessions.id_generator": _id_path,
            "redis.sessions.client_callable": _client_path,
            "redis.sessions.func_invalid_logger": _invalid_logger,
        }
        self.config.include("pyramid_session_redis")
        self.settings = self.config.registry.settings

    def tearDown(self):
        testing.tearDown()

    def test_includeme_serialize_deserialize(self):
        request = get_current_request()  # noqa: F841
        serialize = self.settings["redis.sessions.serialize"]
        deserialize = self.settings["redis.sessions.deserialize"]
        result = deserialize(serialize("test"))
        self.assertEqual(result, "test")

    def test_includeme_id_generator(self):
        request = get_current_request()  # noqa: F841
        generator = self.settings["redis.sessions.id_generator"]
        self.assertEqual(generator(), "42")

    def test_includeme_client_callable(self):
        request = get_current_request()
        get_client = self.settings["redis.sessions.client_callable"]
        self.assertEqual(get_client(request), "client")

    def test_includeme_invalid_logger(self):
        request = get_current_request()  # noqa: F841
        logging_func = self.settings["redis.sessions.func_invalid_logger"]
        raised_error = InvalidSession_DeserializationError("foo")
        # check to ensure this is an InvalidSession instance
        self.assertTrue(logging_func(request, raised_error))


class Test_includeme_advanced(unittest.TestCase):
    def test_fails__no_cookiesigner__no_secret(self):
        request = testing.DummyRequest()
        self.config = testing.setUp(request=request)
        self.config.registry.settings = {
            "redis.sessions.db": 9,
            "redis.sessions.serialize": "pickle.dumps",
            "redis.sessions.deserialize": "pickle.loads",
            "redis.sessions.id_generator": _id_path,
            "redis.sessions.client_callable": _client_path,
            "redis.sessions.func_invalid_logger": _invalid_logger,
            # "redis.sessions.secret": "supersecret",  # don't include!
            # "redis.sessions.cookie_signer": "",  # don't include!
        }
        with self.assertRaises(ConfigurationError) as cm:
            self.config.include("pyramid_session_redis")
        self.assertEqual(
            cm.exception.args[0],
            "One, and only one, of `redis.sessions.secret` and "
            "`redis.sessions.cookie_signer` must be provided.",
        )

    def test_fails__cookiesigner__secret(self):
        request = testing.DummyRequest()
        self.config = testing.setUp(request=request)
        self.config.registry.settings = {
            "redis.sessions.db": 9,
            "redis.sessions.serialize": "pickle.dumps",
            "redis.sessions.deserialize": "pickle.loads",
            "redis.sessions.id_generator": _id_path,
            "redis.sessions.client_callable": _client_path,
            "redis.sessions.func_invalid_logger": _invalid_logger,
            "redis.sessions.secret": "supersecret",
            "redis.sessions.cookie_signer": CustomCookieSigner(),
        }
        with self.assertRaises(ConfigurationError) as cm:
            self.config.include("pyramid_session_redis")
        self.assertEqual(
            cm.exception.args[0],
            "One, and only one, of `redis.sessions.secret` and "
            "`redis.sessions.cookie_signer` must be provided.",
        )

    def test__cookiesigner__custom(self):
        request = testing.DummyRequest()
        self.config = testing.setUp(request=request)
        self.config.registry.settings = {
            "redis.sessions.db": 9,
            "redis.sessions.serialize": "pickle.dumps",
            "redis.sessions.deserialize": "pickle.loads",
            "redis.sessions.id_generator": _id_path,
            "redis.sessions.client_callable": _client_path,
            "redis.sessions.func_invalid_logger": _invalid_logger,
            "redis.sessions.cookie_signer": CustomCookieSigner(),
        }



# -*- coding: utf-8 -*-

# stdlib
import hashlib
import pickle
from secrets import token_hex
from typing import Any
from typing import Awaitable
from typing import Callable
from typing import Dict
from typing import Iterator
from typing import List
from typing import Optional
from typing import overload
from typing import Tuple
from typing import TYPE_CHECKING
from typing import Union

# pypi
from pyramid.decorator import reify
from pyramid.interfaces import ISession
from typing_extensions import Literal
from zope.interface import implementer

# local
from .exceptions import InvalidSession_DeserializationError
from .exceptions import InvalidSession_Lazycreate
from .exceptions import InvalidSession_NotInBackend
from .exceptions import InvalidSession_PayloadLegacy
from .exceptions import InvalidSession_PayloadTimeout
from .exceptions import RawDeserializationError
from .util import decode_session_payload as decode_session_payload_func
from .util import empty_session_payload
from .util import encode_session_payload as encode_session_payload_func
from .util import int_time
from .util import LazyCreateSession
from .util import NotSpecified
from .util import persist
from .util import recookie
from .util import refresh
from .util import SESSION_API_VERSION
from .util import TYPING_COOKIE_EXPIRES
from .util import TYPING_COOKIE_EXPIRES__A
from .util import TYPING_COOKIE_MAX_AGE__A
from .util import TYPING_KEY
from .util import TYPING_SESSION_ID

if TYPE_CHECKING:
    from collections.abc import ItemsView
    from collections.abc import KeysView

    # from pyramid.request import Request  # webob has stubs; pyramid does not
    from redis.client import Redis as RedisClient
    from webob.request import Request

# ==============================================================================


def hashed_value(serialized: bytes) -> str:
    """
    quick hash of serialized data
    only used for comparison

    :param serialized: string. serialized data to hash.
    :returns hash: string.
    """
    return hashlib.md5(serialized).hexdigest()


class _SessionState(object):
    session_id: TYPING_SESSION_ID
    managed_dict: Dict
    created: int
    timeout: Optional[int]
    expires: Optional[int]  # see `util.empty_session_payload`
    version: Optional[int]
    new: bool
    persisted_hash: Optional[str]

    # markers for update
    please_persist: Optional[bool] = None
    please_recookie: Optional[bool] = None
    please_refresh: Optional[bool] = None

    # these markers are consulted in cleanup routines
    dont_persist: Optional[bool] = None
    dont_refresh: Optional[bool] = None

    # optional attributes, not set by default
    cookie_expires: TYPING_COOKIE_EXPIRES__A = (
        NotSpecified  # TYPING_COOKIE_EXPIRES__A includes None
    )
    cookie_max_age: TYPING_COOKIE_MAX_AGE__A = (
        NotSpecified  # TYPING_COOKIE_MAX_AGE__A includes None
    )

    def __init__(
        self,
        session_id: TYPING_SESSION_ID,
        managed_dict: Dict,
        created: int,
        timeout: Optional[int],  # loaded off dict
        expires: Optional[int],  # loaded off dict
        version: Optional[int],  # loaded off dict
        new: bool,
        persisted_hash: Optional[str],
    ):
        """
        all these args are guaranteed to be submitted for the object;
        no need to create default object attributes
        """
        self.session_id = session_id
        self.managed_dict = managed_dict
        self.created = created
        self.timeout = timeout
        self.expires = expires
        self.version = version
        self.new = new
        self.persisted_hash = persisted_hash

    def should_persist(
        self,
        session: "RedisSession",
    ) -> Union[Literal[False], bytes]:
        """
        this is a backup routine
        compares the persisted hash with a hash of the current value
        returns `False` or `serialized_session`

        :param session:
        :returns serialized_session:
        """
        if self.dont_persist:
            return False
        if self.expires and session._timeout_trigger:
            if session.timestamp >= (self.expires - session._timeout_trigger):
                self.please_persist = True
        if self.please_persist:
            return session.to_redis()
        if not session._detect_changes:
            return False
        serialized_session = session.to_redis()
        serialized_hash = hashed_value(serialized_session)
        if serialized_hash == self.persisted_hash:
            return False
        return serialized_session


@implementer(ISession)
class RedisSession(object):
    """
    Implements the Pyramid ISession and IDict interfaces and is returned by
    the ``RedisSessionFactory``.

    Methods that modify the ``dict`` (get, set, update, etc.) are decorated
    with ``@persist`` to update the persisted copy in Redis and reset the
    timeout.

    Methods that are read-only (items, keys, values, etc.) are decorated
    with ``@refresh`` to reset the session's expire time in Redis.

    Methods that request the SetCookie headers are updated are decorated
    with ``@recookie``.

    Session methods make use of the dict methods that already communicate with
    Redis, so they are not decorated.

    Parameters:

    ``redis``
    A Redis connection object.

    ``session_id``
    A unique string associated with the session. Used as a prefix for keys
    and hashes associated with the session.

    ``new``
    Boolean. Whether this session is new (whether it was created in this
    request).

    ``new_session``
    A function that takes no arguments. It should insert a new session into
    Redis under a new session_id, and return that session_id.

    ``new_session_payload``
    UNDER DEVELOPMENT
    A function that takes no arguments.  It is used to to generate a new
    session payload without creating the id (as new_session might)

    ``serialize``
    A function to serialize pickleable Python objects. Default:
    ``cPickle.dumps``.

    ``deserialize``
    The dual of ``serialize``, to convert serialized strings back to Python
    objects. Default: ``cPickle.loads``.

    ``set_redis_ttl``
     If ``True`` sets TTL data in Redis.  If ``False`` assumes Redis is
     configured as a LRU and does not update the expiry data via SETEX.
     Default: ``True``

    ``set_redis_ttl_readheavy``
     If ``True``, sets TTL data in Redis within a PIPELINE via GET+EXPIRE and
     supresses automatic TTL refresh during the deferred cleanup phase. If not
     ``True``, an EXPIRE is sent as a separate action during the deferred
     cleanup phase.  The optimized behavior improves performance on read-heavy
     operations, but may degrade performance on write-heavy operations.  This
     requires a ``timeout`` and ``set_redis_ttl`` to be True; it is not
     compatible with ``timeout_trigger`` or ``python_expires``.
     Default: ``None``

    ``_set_redis_ttl_onexit`` If ``True``, automatically queues a TTL Redis set
    during the cleanup phase. This should be calculated  based on the following
    criteria:
        * self._timeout
        * self._set_redis_ttl
        * not self._timeout_trigger
        * not self._python_expires
        * not self._set_redis_ttl_readheavy
    This is handled as a config option and not a realtime calcluation to save
    some processing. Unit Tests will want to pre-calculate this, otherwise the
    main factory API of this package handles it.
    Default: ``None``

    ``detect_changes``
    If ``True``, supports change detection Default: ``True``

    ``deserialized_fails_new``
    If ``True`` will handle deserializtion errors by creating a new session.

    ``new_payload_func``
    Default ``None``.  Function used to create a new session.

    ``timeout_trigger``
    Default ``None``.  If an int, used to trigger timeouts.

    ``python_expires``
    Default ``None``.  If True, Python is used to manage timeout data.  setting
    ``timeout_trigger`` will enable this.

    """

    redis: "RedisClient"
    _set_redis_ttl: bool
    _set_redis_ttl_readheavy: Optional[bool]
    _detect_changes: bool
    _deserialized_fails_new: Optional[bool]
    _timeout: Optional[int]
    _timeout_trigger: Optional[int]
    _python_expires: Optional[bool]
    _new: bool

    def __init__(
        self,
        redis: "RedisClient",
        session_id: TYPING_SESSION_ID,
        new: bool,
        new_session: Callable[[], TYPING_SESSION_ID],
        new_payload_func: Optional[Callable[..., Dict]] = None,
        serialize: Callable[
            [
                Dict,
            ],
            bytes,
        ] = pickle.dumps,  # dict->bytes
        deserialize: Callable[[bytes], Dict] = pickle.loads,  # bytes->dict
        set_redis_ttl: bool = True,
        detect_changes: bool = True,
        deserialized_fails_new: Optional[bool] = None,
        encode_session_payload_func: Optional[Callable[..., Dict]] = None,
        decode_session_payload_func: Optional[Callable[[Dict], Dict]] = None,
        timeout: Optional[int] = None,
        timeout_trigger: Optional[int] = None,
        python_expires: Optional[bool] = None,
        set_redis_ttl_readheavy: Optional[bool] = None,
        _set_redis_ttl_onexit: Optional[bool] = None,
    ):
        if timeout_trigger and not python_expires:  # fix this
            python_expires = True
        if not timeout:
            if set_redis_ttl_readheavy:
                raise ValueError("`set_redis_ttl_readheavy` requires `timeout`")
        self.redis = redis
        self.serialize = serialize  # type: ignore[method-assign, assignment]
        self.deserialize = deserialize  # type: ignore[method-assign, assignment]
        self.new_session = new_session  # type: ignore[method-assign]
        if new_payload_func is not None:
            self.new_payload = new_payload_func  # type: ignore[method-assign]
        if encode_session_payload_func is not None:
            self.encode_session_payload = encode_session_payload_func  # type: ignore[method-assign]
        if decode_session_payload_func is not None:
            self.decode_session_payload = decode_session_payload_func  # type: ignore[method-assign, assignment]
        self._set_redis_ttl = set_redis_ttl
        self._set_redis_ttl_readheavy = set_redis_ttl_readheavy
        self._detect_changes = detect_changes
        self._deserialized_fails_new = deserialized_fails_new
        self._timeout = timeout
        self._timeout_trigger = timeout_trigger
        self._python_expires = python_expires
        self._new = new
        self._session_state = self._make_session_state(
            session_id=session_id,
            new=new,
        )
        if _set_redis_ttl_onexit:
            self._session_state.please_refresh = True

    def _resync(self) -> None:
        """resyncs the session. this is really only needed for testing."""
        self._session_state = self._make_session_state(
            session_id=self.session_id, new=self._new
        )

    def new_session(self) -> TYPING_SESSION_ID:
        # this should be set via __init__
        raise NotImplementedError()

    def new_payload(self) -> Dict:
        # this should be set via __init__
        return empty_session_payload()

    def encode_session_payload(self, *args, **kwargs):
        """
        used to recode for serialization
        this can be overridden via __init__
        """
        return encode_session_payload_func(*args, **kwargs)

    def decode_session_payload(self, payload: Dict) -> Dict:
        """
        used to recode for serialization
        this can be overridden via __init__

        :param payload:
        :type payload: dict
        :returns decoded payload:
        """
        return decode_session_payload_func(payload)

    def serialize(self, data: Dict) -> bytes:
        # this should be set via __init__
        raise NotImplementedError()

    def deserialize(self, serialized: bytes) -> Dict:
        # this should be set via __init__
        raise NotImplementedError()

    @reify
    def _session_state(self) -> _SessionState:
        """this should only be executed after an `invalidate()`
        The `invalidate()` will "del self._session_state", which will remove
        the '_session_state' entry from the object dict (as created by __init__
        or by this function which reify's the return value). Removing that
        function allows this method to execute, and `reify` puts the new result
        in the object's dict.
        """
        return self._make_session_state(
            session_id=LazyCreateSession,
            new=True,
        )

    @reify
    def timestamp(self) -> int:
        return int_time()

    def _make_session_state(
        self,
        session_id: TYPING_SESSION_ID,
        new: bool,
    ) -> _SessionState:
        """
        This will try to load the session_id in Redis via ``from_redis``.
        If this fails, it will raise ``InvalidSession`` variants

        :param session_id:
        :param new:
        :returns `_SessionState``:
        """
        if session_id is LazyCreateSession:
            persisted_hash = None
            persisted = self.new_payload()
        else:
            # self.from_redis needs to take a session_id here, because
            # otherwise it would look up self.session_id, which is not ready
            # yet as session_state has not been created yet.
            (persisted, persisted_hash) = self.from_redis(
                session_id=session_id,
                persisted_hash=(True if self._detect_changes else False),
            )
            expires = persisted.get("x")
            if expires:
                if self.timestamp > expires:
                    raise InvalidSession_PayloadTimeout(
                        "`session_id` (%s) timed out in python" % session_id
                    )
            version = persisted.get("v")
            if not version or (version < SESSION_API_VERSION):
                raise InvalidSession_PayloadLegacy(
                    "`session_id` (%s) is a legacy format" % session_id
                )
        return _SessionState(
            session_id=session_id,
            managed_dict=persisted["m"],  # managed_dict
            created=persisted["c"],  # created
            timeout=persisted.get("t"),  # timeout
            expires=persisted.get("x"),  # expires
            version=persisted.get("v"),  # session api version
            new=new,
            persisted_hash=persisted_hash,
        )

    @property
    def session_id(self) -> str:
        return self._session_state.session_id

    @property
    def managed_dict(self) -> Dict:
        return self._session_state.managed_dict

    @property
    def created(self) -> int:
        return self._session_state.created

    @property
    def timeout(self) -> int:
        return self._session_state.timeout

    @property
    def expires(self) -> int:
        return self._session_state.expires

    @property
    def version(self) -> int:
        return self._session_state.version

    @property
    def new(self) -> bool:
        return self._session_state.new

    def to_redis(self) -> bytes:
        """Serialize a dict of the data that needs to be persisted for this
        session, for storage in Redis.

        Primarily used by the ``@persist`` decorator to save the current
        session state to Redis.
        """
        data = self.encode_session_payload(
            self.managed_dict,
            self.created,
            self.timeout,
            self.expires,
            timeout_trigger=self._timeout_trigger,
            python_expires=self._python_expires,
        )
        return self.serialize(data)

    @overload
    def from_redis(  # noqa: E704
        self,
        session_id: Optional[TYPING_SESSION_ID] = None,
        persisted_hash: Optional[None] = None,
    ) -> Dict: ...

    @overload
    def from_redis(  # noqa: E704
        self,
        session_id: Optional[TYPING_SESSION_ID] = None,
        persisted_hash: Literal[False] = False,
    ) -> Tuple[Dict, None]: ...

    @overload
    def from_redis(  # noqa: E704
        self,
        session_id: Optional[TYPING_SESSION_ID] = None,
        persisted_hash: Literal[True] = True,
    ) -> Tuple[Dict, str]: ...

    def from_redis(
        self,
        session_id: Optional[TYPING_SESSION_ID] = None,
        persisted_hash: Optional[Literal[True, False]] = None,
    ):
        """
        Get and deserialize the persisted data for this session from Redis.
        If ``persisted_hash`` is ``None`` (default), returns a single
        variable `deserialized`.
        If set to ``True`` or ``False``, returns a tuple.
        """
        _session_id = session_id or self.session_id
        if _session_id is LazyCreateSession:
            raise InvalidSession_Lazycreate(  # noaq: E501
                "`session_id` is LazyCreateSession"
            )
        if TYPE_CHECKING:
            assert isinstance(_session_id, str)

        # optimize a `TTL refresh` under certain conditions
        persisted: Union[Awaitable, bytes, None] = None
        if self._set_redis_ttl_readheavy:
            if TYPE_CHECKING:
                assert self._timeout is not None
            with self.redis.pipeline() as pipe:
                persisted = pipe.get(_session_id)  # type: ignore[assignment]
                _updated = pipe.expire(  # noqa: F841
                    _session_id,
                    self._timeout,
                )
            # mark that we shouldn't refresh
            self._session_state.dont_refresh = True
        else:
            persisted = self.redis.get(_session_id)

        if TYPE_CHECKING:
            assert not isinstance(persisted, Awaitable)

        if persisted is None:
            raise InvalidSession_NotInBackend(
                "`session_id` (%s) not in Redis" % _session_id
            )

        if TYPE_CHECKING:
            assert isinstance(persisted, bytes)

        try:
            deserialized = self.deserialize(persisted)
        except Exception as e:
            if self._deserialized_fails_new:
                raise InvalidSession_DeserializationError(
                    "`session_id` (%s) did not deserialize correctly"
                    % _session_id  # noaq: E501
                )
            raise RawDeserializationError(e)
        # TODO: typing on return can be better?
        if persisted_hash is True:
            # -> tuple[dict, str]
            return (deserialized, hashed_value(persisted))
        elif persisted_hash is False:
            # -> tuple[dict, None]
            return (deserialized, None)
        # -> dict
        return deserialized

    def invalidate(self) -> None:
        """Invalidate the session."""
        if self.session_id != LazyCreateSession:
            self.redis.delete(self.session_id)
        # Delete the self._session_state attribute so that direct access to or
        # indirect access via other methods and properties to .session_id,
        # .managed_dict, .created, .timeout and .new (i.e. anything stored in
        # self._session_state) after this will trigger the creation of a new
        # session with a new session_id via the `_session_state()` reified
        # property.
        del self._session_state

    def ensure_id(self) -> str:
        # this ensures we have a session_id
        if self._session_state.session_id is LazyCreateSession:
            self._session_state.session_id = self.new_session()
        return self._session_state.session_id

    @property
    def session_id_safecheck(self) -> Optional[str]:
        """if we don't have a managed_dict, return None"""
        if not self.managed_dict:
            return None
        return self.ensure_id()

    def do_persist(self, serialized_session=None) -> None:
        """
        Actually and immediately persist to Redis backend
        Only set a timeout in Redis timeout if we have timeouts AND are not
        in LRU mode.
        Note: this package uses StrictRedis(`key, timeout, value`)
              not Redis(`key, value, timeout`)
        """
        self.ensure_id()
        if serialized_session is None:
            serialized_session = self.to_redis()
        serverside_timeout = (
            True
            if ((self.timeout is not None) and (self._set_redis_ttl))
            else False  # noaq: E501
        )
        if serverside_timeout:
            self.redis.setex(self.session_id, self.timeout, serialized_session)
        else:
            self.redis.set(self.session_id, serialized_session)
        self._session_state.please_persist = False
        self._session_state.dont_refresh = True

    def do_refresh(self, force_redis_ttl=None) -> None:
        """
        Actually and immediately refresh the TTL to Redis backend.
        Does nothing if no timeout set (in LRU mode).
        Optional kwarg for developers ``force_redis_ttl`` (default None)
        can be provided to force a new Redis TTL.
        """
        if force_redis_ttl is not None:
            self.redis.expire(self.session_id, force_redis_ttl)
        else:
            if self.timeout is not None:
                if self._set_redis_ttl:
                    # set a TTL if unless we're in LRU mode
                    self.redis.expire(self.session_id, self.timeout)
        self._session_state.please_refresh = False

    # dict modifying methods decorated with @persist
    @persist
    def __delitem__(self, key: TYPING_KEY) -> None:
        del self.managed_dict[key]

    @persist
    def __setitem__(self, key: TYPING_KEY, value: Any) -> None:
        self.managed_dict[key] = value

    @persist
    def setdefault(self, key: TYPING_KEY, default: Optional[Any] = None):
        # TODO: typing
        # this should always return `None`, but mypy doesn't like that
        # unless this should somehow not return `None`
        return self.managed_dict.setdefault(key, default)

    @persist
    def clear(self) -> None:
        return self.managed_dict.clear()

    @persist
    def pop(self, key: TYPING_KEY, default: Optional[Any] = None) -> Any:
        return self.managed_dict.pop(key, default)

    @persist
    def update(self, other) -> None:
        # TODO: typing
        return self.managed_dict.update(other)

    @persist
    def popitem(self) -> tuple:
        # TODO: Deprecate? I can't imagine a usecase for this
        return self.managed_dict.popitem()

    # dict read-only methods decorated with @refresh
    @refresh
    def __getitem__(self, key: str) -> Any:
        return self.managed_dict[key]

    @refresh
    def __contains__(self, key: str) -> bool:
        return key in self.managed_dict

    @refresh
    def keys(self) -> "KeysView":
        # TODO: typing
        return self.managed_dict.keys()

    @refresh
    def items(self) -> "ItemsView":
        # TODO: typing
        return self.managed_dict.items()

    @refresh
    def get(self, key: TYPING_KEY, default: Optional[Any] = None):
        return self.managed_dict.get(key, default)

    @refresh
    def __iter__(self) -> Iterator:
        return self.managed_dict.__iter__()

    @refresh
    def has_key(self, key: TYPING_KEY) -> bool:
        return key in self.managed_dict

    @refresh
    def values(self) -> List:
        return list(self.managed_dict.values())

    @refresh
    def itervalues(self) -> List:
        return list(self.managed_dict.values())

    @refresh
    def iteritems(self) -> List:
        return list(self.managed_dict.items())

    @refresh
    def iterkeys(self) -> List:
        return list(self.managed_dict.keys())

    @persist
    def changed(self) -> None:
        """Persist all the data that needs to be persisted for this session
        immediately with ``@persist``.
        """
        pass

    # session methods persist or refresh using above dict methods
    def new_csrf_token(self) -> str:
        token = token_hex(32)
        self["_csrft_"] = token
        return token

    def get_csrf_token(self) -> str:
        token = self.get("_csrft_", None)
        if token is None:
            token = self.new_csrf_token()
        return token

    def flash(
        self,
        msg: str,
        queue: str = "",
        allow_duplicate: bool = True,
    ) -> None:
        storage = self.setdefault("_f_" + queue, [])
        if allow_duplicate or (msg not in storage):
            storage.append(msg)
            self.changed()  # notify Redis of change to ``storage`` mutable

    def peek_flash(self, queue: str = ""):
        storage = self.get("_f_" + queue, [])
        return storage

    def pop_flash(self, queue: str = ""):
        storage = self.pop("_f_" + queue, [])
        return storage

    # RedisSession extra methods

    @recookie
    def please_recookie(self) -> None:
        """
        this does nothing, other than invoking `@recookie` to ensure the cookie
        is sent with the Response.
        """
        return None

    @recookie
    def adjust_cookie_expires(self, expires: TYPING_COOKIE_EXPIRES) -> None:
        """
        Adjust the `expires` value on the cookie.
        The underlying functionality may be removed in WebOb 1.9.

        This method ONLY affects the SetCookie Headers.
        This method does not affect the session logic or any values.

        A datetime.timedelta object representing an amount of time,
        datetime.datetime or None.

        Expires and Max-Age have a somewhat convoluted relationship;
        Max-Age always takes precedence. You should be using Max-Age instead
        """
        self._session_state.cookie_expires = expires

    @recookie
    def adjust_cookie_max_age(self, max_age: int) -> None:
        """
        Permanently adjusts the max-age for this cookie to ``max_age``
        This value is used as the Max-Age of the generated cookie

        This method ONLY affects the SetCookie Headers.
        This method does not affect the session logic or any values.

        An integer representing a number of seconds, datetime.timedelta,
        or None.

        Expires and Max-Age have a somewhat convoluted relationship;
        Max-Age always takes precedence. You should be using Max-Age.
        """
        self._session_state.cookie_max_age = max_age

    @persist
    def adjust_session_expires(self, expires_epoch: int) -> None:
        """
        Updates the epoch used for Python timeout on expiry logic.
        """
        self._session_state.expires = expires_epoch

    @persist
    def adjust_session_timeout(self, timeout_seconds: int) -> None:
        """
        Permanently adjusts the `timeout` for this Session to
        ``timeout_seconds`` for as long as this Session is active.
        Useful in situations where you want to change the expiry time for
        a Session dynamically.
        """
        self._session_state.timeout = timeout_seconds

    # rename these
    adjust_expires_for_session = adjust_session_expires
    adjust_timeout_for_session = adjust_session_timeout

    @property
    def _invalidated(self) -> bool:
        """
        Boolean property indicating whether the session is in the state where
        it has been invalidated but a new session has not been created in its
        place.
        """
        return "_session_state" not in self.__dict__

    def _deferred_callback(self, request: "Request") -> None:
        """
        Finished callback to persist the data if needed.
        `request` is supplied by pyramid's `add_finished_callback` which should
        invoke this.
        """
        if "_session_state" not in self.__dict__:
            # _session_state is a reified property, which is saved into the
            # dict. if we call `session.invalidate()` the session is immediately
            # deleted however, accessing it here will trigger a new
            # _session_state creation and insert a placeholder for the
            # session_id into Redis.  this would be ok in certain situations,
            # however since we don't access any actual data in the session,
            # it won't have the cookie callback triggered -- which means the
            # cookie will never get sent to the user, and a phantom
            # session_id+placeholder will be in Redis until it times out.
            return
        if not self._session_state.dont_persist:
            if self._session_state.please_persist:
                self.do_persist()
                self._session_state.please_refresh = False
            else:
                if not self.session_id_safecheck:
                    return
                serialized_session = self._session_state.should_persist(self)
                if serialized_session:
                    self.do_persist(serialized_session=serialized_session)
                    self._session_state.please_refresh = False
        if not self._session_state.dont_refresh:
            if self._session_state.please_refresh:
                self.do_refresh()



# -*- coding: utf-8 -*-

# stdlib
from typing import Optional
import unittest

# pypi
from webob.cookies import SignedSerializer

# local
from pyramid_session_redis.exceptions import InvalidSessionId_Serialization
from pyramid_session_redis.legacy import GracefulCookieSerializer
from pyramid_session_redis.legacy import LegacyCookieSerializer
from pyramid_session_redis.legacy import LoggingHookInterface
from pyramid_session_redis.util import _StringSerializer


# ==============================================================================


class TestStringSerializer(unittest.TestCase):
    def test_string(self):
        "this should be a roundtrip"
        serializer = _StringSerializer()
        data_in = "foo"
        data_out = data_in
        _serialized = serializer.dumps(data_in)
        self.assertEqual(data_out, serializer.loads(_serialized))

    def test_int(self):
        "this should fail; _StringSerializer requires a str"
        serializer = _StringSerializer()
        data_in = 100
        self.assertRaises(InvalidSessionId_Serialization, serializer.dumps, data_in)

    def test_bytes(self):
        "this should fail; _StringSerializer requires a str"
        serializer = _StringSerializer()
        data_in = b"foo"
        self.assertRaises(InvalidSessionId_Serialization, serializer.dumps, data_in)


class TestCookieSerialization(unittest.TestCase):
    def _makeOne_default(self, secret: str) -> SignedSerializer:
        # webob docs `SignedSerializer`:
        # An object with two methods: `loads`` and ``dumps``.  The ``loads`` method
        # should accept bytes and return a Python object.  The ``dumps`` method
        # should accept a Python object and return bytes.

        # SignedSerializer(secret, salt, hashalg="sha512", serializer=None)
        cookie_signer = SignedSerializer(
            secret,
            "pyramid_session_redis.",
            "sha512",
            serializer=_StringSerializer(),
        )
        return cookie_signer

    def _makeOne_legacy(self, secret: str) -> LegacyCookieSerializer:
        cookie_signer = LegacyCookieSerializer(secret)
        return cookie_signer

    def _makeOne_graceful(
        self,
        secret: str,
        logging_hook: Optional[LoggingHookInterface] = None,
    ) -> GracefulCookieSerializer:
        cookie_signer = GracefulCookieSerializer(
            secret,
            logging_hook=logging_hook,
        )
        return cookie_signer

    def test_roundtrip_default(self):
        secret = "foo"
        session_id = "session_id:123"
        cookie_signer = self._makeOne_default(secret)
        _serialized = cookie_signer.dumps(session_id)
        self.assertEqual(session_id, cookie_signer.loads(_serialized))

    def test_roundtrip_legacy(self):
        secret = "foo"
        session_id = "session_id:123"
        cookie_signer = self._makeOne_legacy(secret)
        _serialized = cookie_signer.dumps(session_id)
        self.assertEqual(session_id, cookie_signer.loads(_serialized))

    def test_incompatible(self):
        secret = "foo"
        session_id = "session_id:123"
        cookie_signer_current = self._makeOne_default(secret)
        cookie_signer_legacy = self._makeOne_legacy(secret)
        _serialized_current = cookie_signer_current.dumps(session_id)
        _serialized_legacy = cookie_signer_legacy.dumps(session_id)
        self.assertNotEqual(_serialized_current, _serialized_legacy)
        self.assertRaises(
            ValueError, cookie_signer_legacy.loads, _serialized_current
        )  # noqa: E501
        self.assertRaises(
            ValueError, cookie_signer_current.loads, _serialized_legacy
        )  # noqa: E501

    def test_graceful(self):
        secret = "foo"
        session_id = "session_id:123"
        cookie_signer_current = self._makeOne_default(secret)
        cookie_signer_legacy = self._makeOne_legacy(secret)
        cookie_signer_graceful = self._makeOne_graceful(secret)

        _serialized_current = cookie_signer_current.dumps(session_id)
        _serialized_legacy = cookie_signer_legacy.dumps(session_id)
        _serialized_graceful = cookie_signer_graceful.dumps(session_id)

        self.assertEqual(_serialized_current, _serialized_graceful)
        self.assertNotEqual(_serialized_legacy, _serialized_graceful)

        self.assertEqual(
            session_id, cookie_signer_graceful.loads(_serialized_current)
        )  # noqa: E501
        self.assertEqual(
            session_id, cookie_signer_graceful.loads(_serialized_graceful)
        )  # noqa: E501
        self.assertEqual(
            session_id, cookie_signer_graceful.loads(_serialized_legacy)
        )  # noqa: E501

    def test_graceful_hooks(self):
        secret = "foo"
        session_id = "session_id:123"

        class LoggingHook(LoggingHookInterface):
            def __init__(self):
                self._attempts_global = []
                self._attempts = []
                self._successes = []

            def attempt(self, serializer: str) -> None:
                if serializer == "global":
                    self._attempts_global.append(serializer)
                else:
                    self._attempts.append(serializer)

            def success(self, serializer: str) -> None:
                self._successes.append(serializer)

        logging_hook = LoggingHook()

        cookie_signer_legacy = self._makeOne_legacy(secret)
        cookie_signer_graceful = self._makeOne_graceful(secret, logging_hook)
        _serialized_graceful = cookie_signer_graceful.dumps(session_id)
        _serialized_legacy = cookie_signer_legacy.dumps(session_id)

        self.assertEqual(len(logging_hook._attempts), 0)
        self.assertEqual(len(logging_hook._successes), 0)

        cookie_signer_graceful.loads(_serialized_graceful)
        self.assertEqual(len(logging_hook._attempts_global), 1)
        self.assertEqual(len(logging_hook._attempts), 1)
        self.assertEqual(len(logging_hook._successes), 1)

        cookie_signer_graceful.loads(_serialized_legacy)
        self.assertEqual(len(logging_hook._attempts_global), 2)
        self.assertEqual(len(logging_hook._attempts), 3)
        self.assertEqual(len(logging_hook._successes), 2)

        cookie_signer_graceful.loads(_serialized_graceful)
        self.assertEqual(len(logging_hook._attempts_global), 3)
        self.assertEqual(len(logging_hook._attempts), 4)
        self.assertEqual(len(logging_hook._successes), 3)

        self.assertRaises(ValueError, cookie_signer_graceful.loads, "foo")
        self.assertEqual(len(logging_hook._attempts_global), 4)
        self.assertEqual(len(logging_hook._attempts), 6)
        self.assertEqual(len(logging_hook._successes), 3)



# stdlib
from typing import Optional

# pypi
from pyramid.config import Configurator
from pyramid.request import Request
from pyramid.response import Response
from waitress import serve

# local
from ._util import LIVE_PSR_CONFIG

# ==============================================================================


def session_access__none(request: Request) -> Response:
    return Response("<body><h1>session_access__none</h1></body>")


def session_access__set(request: Request) -> Response:
    request.session["a"] = 1
    return Response("<body><h1>session_access__set</h1></body>")


def session_access__set_and_unset(request: Request) -> Response:
    request.session["a"] = 1
    del request.session["a"]
    return Response("<body><h1>session_access__set_and_unset</h1></body>")


def session_access__unset(request: Request) -> Response:
    try:
        del request.session["a"]
    except Exception:
        pass
    return Response("<body><h1>session_access__unset</h1></body>")


def main(global_config: Optional[str], **settings):
    """This function returns a Pyramid WSGI application."""
    config = Configurator(settings=settings)
    config.include("pyramid_session_redis")

    # --------------------------------------------------------------------------

    # session_access__none
    config.add_route("session_access__none", "/session_access__none")
    config.add_view(session_access__none, route_name="session_access__none")

    # session_access__set
    config.add_route("session_access__set", "/session_access__set")
    config.add_view(session_access__set, route_name="session_access__set")

    # session_access__set_and_unset
    config.add_route("session_access__set_and_unset", "/session_access__set_and_unset")
    config.add_view(
        session_access__set_and_unset, route_name="session_access__set_and_unset"
    )

    # session_access__unset
    config.add_route("session_access__unset", "/session_access__unset")
    config.add_view(session_access__unset, route_name="session_access__unset")

    # --------------------------------------------------------------------------

    app = config.make_wsgi_app()
    return app


if __name__ == "__main__":
    app = main(None, **LIVE_PSR_CONFIG)
    serve(app, host="0.0.0.0", port=6543)



# -*- coding: utf-8 -*-

"""
Compatability module for various pythons and environments.
"""

# stdlib
import sys

# This moved in py3.10
if sys.version_info.major == 3:
    if sys.version_info.minor >= 10:
        import collections

        collections.Callable = collections.abc.Callable  # type: ignore[attr-defined]

# ==============================================================================
